#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Common Carla code
# Copyright (C) 2011-2025 Filipe Coelho <falktx@falktx.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# For a full copy of the GNU General Public License see the doc/GPL.txt file.

# ------------------------------------------------------------------------------------------------------------
# Imports (Global)

import os
import sys

from math import fmod

# ------------------------------------------------------------------------------------------------------------
# Imports (Signal)

from signal import signal, SIGINT, SIGTERM

try:
    from signal import SIGUSR1
    haveSIGUSR1 = True
except:
    haveSIGUSR1 = False

# ------------------------------------------------------------------------------------------------------------
# Imports (PyQt5)

# import changed in PyQt 5.15.8, so try both
try:
    from PyQt5.Qt import PYQT_VERSION_STR
except ImportError:
    from PyQt5.QtCore import PYQT_VERSION_STR

from PyQt5.QtCore import qFatal, QT_VERSION, QT_VERSION_STR, qWarning, QDir, QSettings
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QFileDialog, QMessageBox

# ------------------------------------------------------------------------------------------------------------
# Imports (Custom)

from carla_backend import (
    kIs64bit, HAIKU, LINUX, MACOS, WINDOWS,
    MAX_DEFAULT_PARAMETERS,
    ENGINE_PROCESS_MODE_MULTIPLE_CLIENTS,
    ENGINE_PROCESS_MODE_PATCHBAY,
    ENGINE_TRANSPORT_MODE_INTERNAL,
    ENGINE_TRANSPORT_MODE_JACK
)

# ------------------------------------------------------------------------------------------------------------
# Config

# These will be modified during install
X_LIBDIR_X = "/usr/lib"
X_DATADIR_X = "/usr/share"

# ------------------------------------------------------------------------------------------------------------
# Platform specific stuff

if WINDOWS:
    WINDIR = os.getenv("WINDIR")

# ------------------------------------------------------------------------------------------------------------
# Set Version

VERSION = "2.5.10"

# ------------------------------------------------------------------------------------------------------------
# Set TMP

envTMP = os.getenv("TMP")

if envTMP is None:
    if WINDOWS:
        qWarning("TMP variable not set")
    TMP = QDir.tempPath()
else:
    TMP = envTMP

if not os.path.exists(TMP):
    qWarning("TMP does not exist")
    TMP = "/"

del envTMP

# ------------------------------------------------------------------------------------------------------------
# Set HOME

envHOME = os.getenv("HOME")

if envHOME is None:
    if LINUX or MACOS:
        qWarning("HOME variable not set")
    HOME = QDir.toNativeSeparators(QDir.homePath())
else:
    HOME = envHOME

if not os.path.exists(HOME):
    qWarning("HOME does not exist")
    HOME = TMP

del envHOME

# ------------------------------------------------------------------------------------------------------------
# Set PATH

envPATH = os.getenv("PATH")

if envPATH is None:
    qWarning("PATH variable not set")
    if MACOS:
        PATH = ("/opt/local/bin", "/usr/local/bin", "/usr/bin", "/bin")
    elif WINDOWS:
        PATH = (os.path.join(WINDIR, "system32"), WINDIR)
    else:
        PATH = ("/usr/local/bin", "/usr/bin", "/bin")
else:
    PATH = envPATH.split(os.pathsep)

del envPATH

# ------------------------------------------------------------------------------------------------------------
# Static MIDI CC list

MIDI_CC_LIST = (
    "01 [0x01] Modulation",
    "02 [0x02] Breath",
    "04 [0x04] Foot",
    "05 [0x05] Portamento",
    "07 [0x07] Volume",
    "08 [0x08] Balance",
    "10 [0x0A] Pan",
    "11 [0x0B] Expression",
    "12 [0x0C] FX Control 1",
    "13 [0x0D] FX Control 2",
    "16 [0x10] General Purpose 1",
    "17 [0x11] General Purpose 2",
    "18 [0x12] General Purpose 3",
    "19 [0x13] General Purpose 4",
    "70 [0x46] Control 1 [Variation]",
    "71 [0x47] Control 2 [Timbre]",
    "72 [0x48] Control 3 [Release]",
    "73 [0x49] Control 4 [Attack]",
    "74 [0x4A] Control 5 [Brightness]",
    "75 [0x4B] Control 6 [Decay]",
    "76 [0x4C] Control 7 [Vib Rate]",
    "77 [0x4D] Control 8 [Vib Depth]",
    "78 [0x4E] Control 9 [Vib Delay]",
    "79 [0x4F] Control 10 [Undefined]",
    "80 [0x50] General Purpose 5",
    "81 [0x51] General Purpose 6",
    "82 [0x52] General Purpose 7",
    "83 [0x53] General Purpose 8",
    "84 [0x54] Portamento Control",
    "91 [0x5B] FX 1 Depth [Reverb]",
    "92 [0x5C] FX 2 Depth [Tremolo]",
    "93 [0x5D] FX 3 Depth [Chorus]",
    "94 [0x5E] FX 4 Depth [Detune]",
    "95 [0x5F] FX 5 Depth [Phaser]"
)

MAX_MIDI_CC_LIST_ITEM = 95

# ------------------------------------------------------------------------------------------------------------
# PatchCanvas defines

CANVAS_ANTIALIASING_SMALL = 1
CANVAS_EYECANDY_SMALL     = 1

# ------------------------------------------------------------------------------------------------------------
# Carla Settings keys

CARLA_KEY_MAIN_PROJECT_FOLDER   = "Main/ProjectFolder"   # str
CARLA_KEY_MAIN_USE_PRO_THEME    = "Main/UseProTheme"     # bool
CARLA_KEY_MAIN_PRO_THEME_COLOR  = "Main/ProThemeColor"   # str
CARLA_KEY_MAIN_REFRESH_INTERVAL = "Main/RefreshInterval" # int
CARLA_KEY_MAIN_CONFIRM_EXIT     = "Main/ConfirmExit"     # bool
CARLA_KEY_MAIN_CLASSIC_SKIN     = "Main/ClassicSkin"     # bool
CARLA_KEY_MAIN_SHOW_LOGS        = "Main/ShowLogs"        # bool
CARLA_KEY_MAIN_SYSTEM_ICONS     = "Main/SystemIcons"     # bool
CARLA_KEY_MAIN_EXPERIMENTAL     = "Main/Experimental"    # bool

CARLA_KEY_CANVAS_THEME             = "Canvas/Theme"           # str
CARLA_KEY_CANVAS_SIZE              = "Canvas/Size"            # str "NxN"
CARLA_KEY_CANVAS_USE_BEZIER_LINES  = "Canvas/UseBezierLines"  # bool
CARLA_KEY_CANVAS_AUTO_HIDE_GROUPS  = "Canvas/AutoHideGroups"  # bool
CARLA_KEY_CANVAS_AUTO_SELECT_ITEMS = "Canvas/AutoSelectItems" # bool
CARLA_KEY_CANVAS_EYE_CANDY         = "Canvas/EyeCandy2"       # bool
CARLA_KEY_CANVAS_FANCY_EYE_CANDY   = "Canvas/FancyEyeCandy"   # bool
CARLA_KEY_CANVAS_USE_OPENGL        = "Canvas/UseOpenGL"       # bool
CARLA_KEY_CANVAS_ANTIALIASING      = "Canvas/Antialiasing"    # enum
CARLA_KEY_CANVAS_HQ_ANTIALIASING   = "Canvas/HQAntialiasing"  # bool
CARLA_KEY_CANVAS_INLINE_DISPLAYS   = "Canvas/InlineDisplays"  # bool
CARLA_KEY_CANVAS_FULL_REPAINTS     = "Canvas/FullRepaints"    # bool

CARLA_KEY_ENGINE_DRIVER_PREFIX         = "Engine/Driver-"
CARLA_KEY_ENGINE_AUDIO_DRIVER          = "Engine/AudioDriver"         # str
CARLA_KEY_ENGINE_PROCESS_MODE          = "Engine/ProcessMode"         # enum
CARLA_KEY_ENGINE_TRANSPORT_MODE        = "Engine/TransportMode"       # enum
CARLA_KEY_ENGINE_TRANSPORT_EXTRA       = "Engine/TransportExtra"      # str
CARLA_KEY_ENGINE_FORCE_STEREO          = "Engine/ForceStereo"         # bool
CARLA_KEY_ENGINE_PREFER_PLUGIN_BRIDGES = "Engine/PreferPluginBridges" # bool
CARLA_KEY_ENGINE_PREFER_UI_BRIDGES     = "Engine/PreferUiBridges"     # bool
CARLA_KEY_ENGINE_MANAGE_UIS            = "Engine/ManageUIs"           # bool
CARLA_KEY_ENGINE_UIS_ALWAYS_ON_TOP     = "Engine/UIsAlwaysOnTop"      # bool
CARLA_KEY_ENGINE_MAX_PARAMETERS        = "Engine/MaxParameters"       # int
CARLA_KEY_ENGINE_RESET_XRUNS           = "Engine/ResetXruns"          # bool
CARLA_KEY_ENGINE_UI_BRIDGES_TIMEOUT    = "Engine/UiBridgesTimeout"    # int

CARLA_KEY_OSC_ENABLED          = "OSC/Enabled"
CARLA_KEY_OSC_TCP_PORT_ENABLED = "OSC/TCPEnabled"
CARLA_KEY_OSC_TCP_PORT_NUMBER  = "OSC/TCPNumber"
CARLA_KEY_OSC_TCP_PORT_RANDOM  = "OSC/TCPRandom"
CARLA_KEY_OSC_UDP_PORT_ENABLED = "OSC/UDPEnabled"
CARLA_KEY_OSC_UDP_PORT_NUMBER  = "OSC/UDPNumber"
CARLA_KEY_OSC_UDP_PORT_RANDOM  = "OSC/UDPRandom"

CARLA_KEY_PATHS_AUDIO = "Paths/Audio"
CARLA_KEY_PATHS_MIDI  = "Paths/MIDI"

CARLA_KEY_PATHS_LADSPA = "Paths/LADSPA"
CARLA_KEY_PATHS_DSSI   = "Paths/DSSI"
CARLA_KEY_PATHS_LV2    = "Paths/LV2"
CARLA_KEY_PATHS_VST2   = "Paths/VST2"
CARLA_KEY_PATHS_VST3   = "Paths/VST3"
CARLA_KEY_PATHS_SF2    = "Paths/SF2"
CARLA_KEY_PATHS_SFZ    = "Paths/SFZ"
CARLA_KEY_PATHS_JSFX   = "Paths/JSFX"

CARLA_KEY_WINE_EXECUTABLE      = "Wine/Executable"     # str
CARLA_KEY_WINE_AUTO_PREFIX     = "Wine/AutoPrefix"     # bool
CARLA_KEY_WINE_FALLBACK_PREFIX = "Wine/FallbackPrefix" # str
CARLA_KEY_WINE_RT_PRIO_ENABLED = "Wine/RtPrioEnabled"  # bool
CARLA_KEY_WINE_BASE_RT_PRIO    = "Wine/BaseRtPrio"     # int
CARLA_KEY_WINE_SERVER_RT_PRIO  = "Wine/ServerRtPrio"   # int

CARLA_KEY_EXPERIMENTAL_PLUGIN_BRIDGES        = "Experimental/PluginBridges"       # bool
CARLA_KEY_EXPERIMENTAL_WINE_BRIDGES          = "Experimental/WineBridges"         # bool
CARLA_KEY_EXPERIMENTAL_JACK_APPS             = "Experimental/JackApplications"    # bool
CARLA_KEY_EXPERIMENTAL_EXPORT_LV2            = "Experimental/ExportLV2"           # bool
CARLA_KEY_EXPERIMENTAL_PREVENT_BAD_BEHAVIOUR = "Experimental/PreventBadBehaviour" # bool
CARLA_KEY_EXPERIMENTAL_LOAD_LIB_GLOBAL       = "Experimental/LoadLibGlobal"       # bool

# if pro theme is on and color is black
CARLA_KEY_CUSTOM_PAINTING = "UseCustomPainting" # bool

# ------------------------------------------------------------------------------------------------------------
# Carla Settings defaults

# Main
CARLA_DEFAULT_MAIN_PROJECT_FOLDER   = HOME
CARLA_DEFAULT_MAIN_USE_PRO_THEME    = True
CARLA_DEFAULT_MAIN_PRO_THEME_COLOR  = "Black"
CARLA_DEFAULT_MAIN_REFRESH_INTERVAL = 20
CARLA_DEFAULT_MAIN_CONFIRM_EXIT     = True
CARLA_DEFAULT_MAIN_CLASSIC_SKIN     = False
CARLA_DEFAULT_MAIN_SHOW_LOGS        = bool(not WINDOWS)
CARLA_DEFAULT_MAIN_SYSTEM_ICONS     = False
CARLA_DEFAULT_MAIN_EXPERIMENTAL     = False

# Canvas
CARLA_DEFAULT_CANVAS_THEME             = "Modern Dark"
CARLA_DEFAULT_CANVAS_SIZE              = "3100x2400"
CARLA_DEFAULT_CANVAS_SIZE_WIDTH        = 3100
CARLA_DEFAULT_CANVAS_SIZE_HEIGHT       = 2400
CARLA_DEFAULT_CANVAS_USE_BEZIER_LINES  = True
CARLA_DEFAULT_CANVAS_AUTO_HIDE_GROUPS  = True
CARLA_DEFAULT_CANVAS_AUTO_SELECT_ITEMS = False
CARLA_DEFAULT_CANVAS_EYE_CANDY         = True
CARLA_DEFAULT_CANVAS_FANCY_EYE_CANDY   = False
CARLA_DEFAULT_CANVAS_USE_OPENGL        = False
CARLA_DEFAULT_CANVAS_ANTIALIASING      = CANVAS_ANTIALIASING_SMALL
CARLA_DEFAULT_CANVAS_HQ_ANTIALIASING   = False
CARLA_DEFAULT_CANVAS_INLINE_DISPLAYS   = False
CARLA_DEFAULT_CANVAS_FULL_REPAINTS     = False

# Engine
CARLA_DEFAULT_FORCE_STEREO          = False
CARLA_DEFAULT_PREFER_PLUGIN_BRIDGES = False
CARLA_DEFAULT_PREFER_UI_BRIDGES     = True
CARLA_DEFAULT_MANAGE_UIS            = True
CARLA_DEFAULT_UIS_ALWAYS_ON_TOP     = False
CARLA_DEFAULT_MAX_PARAMETERS        = MAX_DEFAULT_PARAMETERS
CARLA_DEFAULT_RESET_XRUNS           = False
CARLA_DEFAULT_UI_BRIDGES_TIMEOUT    = 4000

CARLA_DEFAULT_AUDIO_BUFFER_SIZE     = 512
CARLA_DEFAULT_AUDIO_SAMPLE_RATE     = 44100
CARLA_DEFAULT_AUDIO_TRIPLE_BUFFER   = False

if HAIKU:
    CARLA_DEFAULT_AUDIO_DRIVER = "SDL"
elif MACOS:
    CARLA_DEFAULT_AUDIO_DRIVER = "CoreAudio"
elif WINDOWS:
    CARLA_DEFAULT_AUDIO_DRIVER = "Windows Audio"
elif os.path.exists("/usr/bin/jackd") or os.path.exists("/usr/bin/jackdbus") or os.path.exists("/usr/bin/pw-jack"):
    CARLA_DEFAULT_AUDIO_DRIVER = "JACK"
else:
    CARLA_DEFAULT_AUDIO_DRIVER = "PulseAudio"

if CARLA_DEFAULT_AUDIO_DRIVER == "JACK":
    CARLA_DEFAULT_PROCESS_MODE   = ENGINE_PROCESS_MODE_MULTIPLE_CLIENTS
    CARLA_DEFAULT_TRANSPORT_MODE = ENGINE_TRANSPORT_MODE_JACK
else:
    CARLA_DEFAULT_PROCESS_MODE   = ENGINE_PROCESS_MODE_PATCHBAY
    CARLA_DEFAULT_TRANSPORT_MODE = ENGINE_TRANSPORT_MODE_INTERNAL

# OSC
CARLA_DEFAULT_OSC_ENABLED = not (MACOS or WINDOWS)
CARLA_DEFAULT_OSC_TCP_PORT_ENABLED = True
CARLA_DEFAULT_OSC_TCP_PORT_NUMBER  = 22752
CARLA_DEFAULT_OSC_TCP_PORT_RANDOM  = False
CARLA_DEFAULT_OSC_UDP_PORT_ENABLED = True
CARLA_DEFAULT_OSC_UDP_PORT_NUMBER  = 22752
CARLA_DEFAULT_OSC_UDP_PORT_RANDOM  = False

# Wine
CARLA_DEFAULT_WINE_EXECUTABLE      = "wine"
CARLA_DEFAULT_WINE_AUTO_PREFIX     = True
CARLA_DEFAULT_WINE_FALLBACK_PREFIX = os.path.expanduser("~/.wine")
CARLA_DEFAULT_WINE_RT_PRIO_ENABLED = True
CARLA_DEFAULT_WINE_BASE_RT_PRIO    = 15
CARLA_DEFAULT_WINE_SERVER_RT_PRIO  = 10

# Experimental
CARLA_DEFAULT_EXPERIMENTAL_PLUGIN_BRIDGES        = False
CARLA_DEFAULT_EXPERIMENTAL_WINE_BRIDGES          = False
CARLA_DEFAULT_EXPERIMENTAL_JACK_APPS             = False
CARLA_DEFAULT_EXPERIMENTAL_LV2_EXPORT            = False
CARLA_DEFAULT_EXPERIMENTAL_PREVENT_BAD_BEHAVIOUR = False
CARLA_DEFAULT_EXPERIMENTAL_LOAD_LIB_GLOBAL       = False

# ------------------------------------------------------------------------------------------------------------
# Default File Folders

CARLA_DEFAULT_FILE_PATH_AUDIO = []
CARLA_DEFAULT_FILE_PATH_MIDI  = []

# ------------------------------------------------------------------------------------------------------------
# Default Plugin Folders (get)

DEFAULT_LADSPA_PATH = ""
DEFAULT_DSSI_PATH   = ""
DEFAULT_LV2_PATH    = ""
DEFAULT_VST2_PATH   = ""
DEFAULT_VST3_PATH   = ""
DEFAULT_SF2_PATH    = ""
DEFAULT_SFZ_PATH    = ""
DEFAULT_JSFX_PATH   = ""

if WINDOWS:
    splitter = ";"

    APPDATA = os.getenv("APPDATA")
    LOCALAPPDATA = os.getenv("LOCALAPPDATA", APPDATA)
    PROGRAMFILES = os.getenv("PROGRAMFILES")
    PROGRAMFILESx86 = os.getenv("PROGRAMFILES(x86)")
    COMMONPROGRAMFILES = os.getenv("COMMONPROGRAMFILES")
    COMMONPROGRAMFILESx86 = os.getenv("COMMONPROGRAMFILES(x86)")

    # Small integrity tests
    if not APPDATA:
        qFatal("APPDATA variable not set, cannot continue")
        sys.exit(1)

    if not PROGRAMFILES:
        qFatal("PROGRAMFILES variable not set, cannot continue")
        sys.exit(1)

    if not COMMONPROGRAMFILES:
        qFatal("COMMONPROGRAMFILES variable not set, cannot continue")
        sys.exit(1)

    DEFAULT_LADSPA_PATH  = APPDATA + "\\LADSPA"
    DEFAULT_LADSPA_PATH += ";" + PROGRAMFILES + "\\LADSPA"

    DEFAULT_DSSI_PATH    = APPDATA + "\\DSSI"
    DEFAULT_DSSI_PATH   += ";" + PROGRAMFILES + "\\DSSI"

    DEFAULT_LV2_PATH     = APPDATA + "\\LV2"
    DEFAULT_LV2_PATH    += ";" + COMMONPROGRAMFILES + "\\LV2"

    DEFAULT_VST2_PATH    = PROGRAMFILES + "\\VstPlugins"
    DEFAULT_VST2_PATH   += ";" + PROGRAMFILES + "\\Steinberg\\VstPlugins"

    DEFAULT_JSFX_PATH    = APPDATA + "\\REAPER\\Effects"
    #DEFAULT_JSFX_PATH   += ";" + PROGRAMFILES + "\\REAPER\\InstallData\\Effects"

    if kIs64bit:
        DEFAULT_VST2_PATH  += ";" + COMMONPROGRAMFILES + "\\VST2"

    DEFAULT_VST3_PATH    = COMMONPROGRAMFILES + "\\VST3"
    DEFAULT_VST3_PATH   += ";" + LOCALAPPDATA + "\\Programs\\Common\\VST3"

    DEFAULT_SF2_PATH     = APPDATA + "\\SF2"
    DEFAULT_SFZ_PATH     = APPDATA + "\\SFZ"

    if PROGRAMFILESx86:
        DEFAULT_LADSPA_PATH += ";" + PROGRAMFILESx86 + "\\LADSPA"
        DEFAULT_DSSI_PATH   += ";" + PROGRAMFILESx86 + "\\DSSI"
        DEFAULT_VST2_PATH   += ";" + PROGRAMFILESx86 + "\\VstPlugins"
        DEFAULT_VST2_PATH   += ";" + PROGRAMFILESx86 + "\\Steinberg\\VstPlugins"
        #DEFAULT_JSFX_PATH   += ";" + PROGRAMFILESx86 + "\\REAPER\\InstallData\\Effects"

    if COMMONPROGRAMFILESx86:
        DEFAULT_VST3_PATH   += COMMONPROGRAMFILESx86 + "\\VST3"

elif HAIKU:
    splitter = ":"

    DEFAULT_LADSPA_PATH  = HOME + "/.ladspa"
    DEFAULT_LADSPA_PATH += ":/system/add-ons/media/ladspaplugins"
    DEFAULT_LADSPA_PATH += ":/system/lib/ladspa"

    DEFAULT_DSSI_PATH    = HOME + "/.dssi"
    DEFAULT_DSSI_PATH   += ":/system/add-ons/media/dssiplugins"
    DEFAULT_DSSI_PATH   += ":/system/lib/dssi"

    DEFAULT_LV2_PATH     = HOME + "/.lv2"
    DEFAULT_LV2_PATH    += ":/system/add-ons/media/lv2plugins"

    DEFAULT_VST2_PATH    = HOME + "/.vst"
    DEFAULT_VST2_PATH   += ":/system/add-ons/media/vstplugins"

    DEFAULT_VST3_PATH    = HOME + "/.vst3"
    DEFAULT_VST3_PATH   += ":/system/add-ons/media/vst3plugins"

elif MACOS:
    splitter = ":"

    DEFAULT_LADSPA_PATH  = HOME + "/Library/Audio/Plug-Ins/LADSPA"
    DEFAULT_LADSPA_PATH += ":/Library/Audio/Plug-Ins/LADSPA"

    DEFAULT_DSSI_PATH    = HOME + "/Library/Audio/Plug-Ins/DSSI"
    DEFAULT_DSSI_PATH   += ":/Library/Audio/Plug-Ins/DSSI"

    DEFAULT_LV2_PATH     = HOME + "/Library/Audio/Plug-Ins/LV2"
    DEFAULT_LV2_PATH    += ":/Library/Audio/Plug-Ins/LV2"

    DEFAULT_VST2_PATH    = HOME + "/Library/Audio/Plug-Ins/VST"
    DEFAULT_VST2_PATH   += ":/Library/Audio/Plug-Ins/VST"

    DEFAULT_VST3_PATH    = HOME + "/Library/Audio/Plug-Ins/VST3"
    DEFAULT_VST3_PATH   += ":/Library/Audio/Plug-Ins/VST3"

    DEFAULT_JSFX_PATH    = HOME + "/Library/Application Support/REAPER/Effects"
    #DEFAULT_JSFX_PATH   += ":/Applications/REAPER.app/Contents/InstallFiles/Effects"

else:
    CONFIG_HOME = os.getenv("XDG_CONFIG_HOME", HOME + "/.config")

    splitter = ":"

    DEFAULT_LADSPA_PATH  = HOME + "/.ladspa"
    DEFAULT_LADSPA_PATH += ":/usr/lib/ladspa"
    DEFAULT_LADSPA_PATH += ":/usr/local/lib/ladspa"

    DEFAULT_DSSI_PATH    = HOME + "/.dssi"
    DEFAULT_DSSI_PATH   += ":/usr/lib/dssi"
    DEFAULT_DSSI_PATH   += ":/usr/local/lib/dssi"

    DEFAULT_LV2_PATH     = HOME + "/.lv2"
    DEFAULT_LV2_PATH    += ":/usr/lib/lv2"
    DEFAULT_LV2_PATH    += ":/usr/local/lib/lv2"

    DEFAULT_VST2_PATH    = HOME + "/.vst"
    DEFAULT_VST2_PATH   += ":/usr/lib/vst"
    DEFAULT_VST2_PATH   += ":/usr/local/lib/vst"

    DEFAULT_VST2_PATH   += HOME + "/.lxvst"
    DEFAULT_VST2_PATH   += ":/usr/lib/lxvst"
    DEFAULT_VST2_PATH   += ":/usr/local/lib/lxvst"

    DEFAULT_VST3_PATH    = HOME + "/.vst3"
    DEFAULT_VST3_PATH   += ":/usr/lib/vst3"
    DEFAULT_VST3_PATH   += ":/usr/local/lib/vst3"

    DEFAULT_SF2_PATH     = HOME + "/.sounds/sf2"
    DEFAULT_SF2_PATH    += ":" + HOME + "/.sounds/sf3"
    DEFAULT_SF2_PATH    += ":/usr/share/sounds/sf2"
    DEFAULT_SF2_PATH    += ":/usr/share/sounds/sf3"
    DEFAULT_SF2_PATH    += ":/usr/share/soundfonts"

    DEFAULT_SFZ_PATH     = HOME + "/.sounds/sfz"
    DEFAULT_SFZ_PATH    += ":/usr/share/sounds/sfz"

    DEFAULT_JSFX_PATH    = CONFIG_HOME + "/REAPER/Effects"
    #DEFAULT_JSFX_PATH   += ":" + "/opt/REAPER/InstallData/Effects"

if not WINDOWS:
    winePrefix = os.getenv("WINEPREFIX")

    if not winePrefix:
        winePrefix = HOME + "/.wine"

    if os.path.exists(winePrefix):
        DEFAULT_VST2_PATH += ":" + winePrefix + "/drive_c/Program Files/VstPlugins"
        DEFAULT_VST3_PATH += ":" + winePrefix + "/drive_c/Program Files/Common Files/VST3"

        if kIs64bit and os.path.exists(winePrefix + "/drive_c/Program Files (x86)"):
            DEFAULT_VST2_PATH += ":" + winePrefix + "/drive_c/Program Files (x86)/VstPlugins"
            DEFAULT_VST3_PATH += ":" + winePrefix + "/drive_c/Program Files (x86)/Common Files/VST3"

    del winePrefix

# ------------------------------------------------------------------------------------------------------------
# Default Plugin Folders (set)

readEnvVars = True

if WINDOWS:
    # Check if running Wine. If yes, ignore env vars
    # pylint: disable=import-error
    from winreg import ConnectRegistry, OpenKey, CloseKey, HKEY_CURRENT_USER
    # pylint: enable=import-error
    _reg = ConnectRegistry(None, HKEY_CURRENT_USER)

    try:
        _key = OpenKey(_reg, r"SOFTWARE\Wine")
        CloseKey(_key)
        del _key
        readEnvVars = False
    except:
        pass

    CloseKey(_reg)
    del _reg

if readEnvVars:
    CARLA_DEFAULT_LADSPA_PATH = os.getenv("LADSPA_PATH", DEFAULT_LADSPA_PATH).split(splitter)
    CARLA_DEFAULT_DSSI_PATH   = os.getenv("DSSI_PATH",   DEFAULT_DSSI_PATH).split(splitter)
    CARLA_DEFAULT_LV2_PATH    = os.getenv("LV2_PATH",    DEFAULT_LV2_PATH).split(splitter)
    CARLA_DEFAULT_VST2_PATH   = os.getenv("VST_PATH",    DEFAULT_VST2_PATH).split(splitter)
    CARLA_DEFAULT_VST3_PATH   = os.getenv("VST3_PATH",   DEFAULT_VST3_PATH).split(splitter)
    CARLA_DEFAULT_SF2_PATH    = os.getenv("SF2_PATH",    DEFAULT_SF2_PATH).split(splitter)
    CARLA_DEFAULT_SFZ_PATH    = os.getenv("SFZ_PATH",    DEFAULT_SFZ_PATH).split(splitter)
    CARLA_DEFAULT_JSFX_PATH   = os.getenv("JSFX_PATH",   DEFAULT_JSFX_PATH).split(splitter)

else:
    CARLA_DEFAULT_LADSPA_PATH = DEFAULT_LADSPA_PATH.split(splitter)
    CARLA_DEFAULT_DSSI_PATH   = DEFAULT_DSSI_PATH.split(splitter)
    CARLA_DEFAULT_LV2_PATH    = DEFAULT_LV2_PATH.split(splitter)
    CARLA_DEFAULT_VST2_PATH   = DEFAULT_VST2_PATH.split(splitter)
    CARLA_DEFAULT_VST3_PATH   = DEFAULT_VST3_PATH.split(splitter)
    CARLA_DEFAULT_SF2_PATH    = DEFAULT_SF2_PATH.split(splitter)
    CARLA_DEFAULT_SFZ_PATH    = DEFAULT_SFZ_PATH.split(splitter)
    CARLA_DEFAULT_JSFX_PATH   = DEFAULT_JSFX_PATH.split(splitter)

# ------------------------------------------------------------------------------------------------------------
# Default Plugin Folders (cleanup)

del DEFAULT_LADSPA_PATH
del DEFAULT_DSSI_PATH
del DEFAULT_LV2_PATH
del DEFAULT_VST2_PATH
del DEFAULT_VST3_PATH
del DEFAULT_SF2_PATH
del DEFAULT_SFZ_PATH

# ------------------------------------------------------------------------------------------------------------
# Global Carla object

class CarlaObject():
    def __init__(self):
        self.cnprefix = ""    # Client name prefix
        self.gui      = None  # Host Window
        self.nogui    = False # Skip UI
        self.term     = False # Terminated by OS signal
        self.utils    = None  # Utils object

gCarla = CarlaObject()

# ------------------------------------------------------------------------------------------------------------
# Set CWD

CWD = sys.path[0]

if not CWD:
    CWD = os.path.dirname(sys.argv[0])

# make it work with cxfreeze
if os.path.isfile(CWD):
    CWD = os.path.dirname(CWD)
    if CWD.endswith("/lib"):
        CWD = CWD.rsplit("/lib",1)[0]
    CXFREEZE = True
    if not WINDOWS:
        os.environ['CARLA_MAGIC_FILE'] = os.path.join(CWD, "magic.mgc")
else:
    CXFREEZE = False

# ------------------------------------------------------------------------------------------------------------
# Set DLL_EXTENSION

if WINDOWS:
    DLL_EXTENSION = "dll"
elif MACOS:
    DLL_EXTENSION = "dylib"
else:
    DLL_EXTENSION = "so"

# ------------------------------------------------------------------------------------------------------------
# Find decimal points for a parameter, using step and stepSmall

def countDecimalPoints(step, stepSmall):
    if stepSmall >= 1.0:
        return 0
    if step >= 1.0:
        return 2

    count = 0
    value = fmod(abs(step), 1)
    while 0.0001 < value < 0.999 and count < 6:
        value = fmod(value*10, 1)
        count += 1

    return count

# ------------------------------------------------------------------------------------------------------------
# Check if a value is a number (float support)

def isNumber(value):
    try:
        float(value)
        return True
    except:
        return False

# ------------------------------------------------------------------------------------------------------------
# Convert a value to a list

def toList(value):
    if value is None:
        return []
    if not isinstance(value, list):
        return [value]
    return value

# ------------------------------------------------------------------------------------------------------------
# Get Icon from user theme, using our own as backup (Oxygen)

def getIcon(icon, size, qrcformat):
    return QIcon.fromTheme(icon, QIcon(":/%ix%i/%s.%s" % (size, size, icon, qrcformat)))

# ------------------------------------------------------------------------------------------------------------
# Handle some basic command-line arguments shared between all carla variants

def handleInitialCommandLineArguments(file):
    initName  = os.path.basename(file) if (file is not None and os.path.dirname(file) in PATH) else sys.argv[0]
    libPrefix = None
    readPrefixNext = False

    for arg in sys.argv[1:]:
        if arg.startswith("--with-appname="):
            initName = os.path.basename(arg.replace("--with-appname=", ""))

        elif arg.startswith("--with-libprefix="):
            libPrefix = arg.replace("--with-libprefix=", "")

        elif arg.startswith("--osc-gui="):
            gCarla.nogui = int(arg.replace("--osc-gui=", ""))

        elif arg.startswith("--cnprefix="):
            gCarla.cnprefix = arg.replace("--cnprefix=", "")

        elif arg == "--cnprefix":
            readPrefixNext = True

        elif arg == "--gdb":
            pass

        elif arg in ("-n", "--n", "-no-gui", "--no-gui", "-nogui", "--nogui"):
            gCarla.nogui = True

        elif MACOS and arg.startswith("-psn_"):
            pass

        elif arg in ("-h", "--h", "-help", "--help"):
            print("Usage: %s [OPTION]... [FILE|URL]" % initName)
            print("")
            print(" where FILE can be a Carla project or preset file to be loaded, or URL if using Carla-Control")
            print("")
            print(" and OPTION can be one or more of the following:")
            print("")
            print("    --cnprefix\t Set a prefix for client names in multi-client mode.")
            if isinstance(gCarla.nogui, bool):
                if X_LIBDIR_X is not None:
                    print("    --gdb     \t Run Carla inside gdb.")
                print(" -n,--no-gui  \t Run Carla headless, don't show UI.")
                print("")
            print(" -h,--help    \t Print this help text and exit.")
            print(" -v,--version \t Print version information and exit.")
            print("")

            if not isinstance(gCarla.nogui, bool):
                print("NOTE: when using %s the FILE is only valid the first time the backend is started" % initName)
                sys.exit(1)

            sys.exit(0)

        elif arg in ("-v", "--v", "-version", "--version"):
            pathBinaries, pathResources = getPaths(libPrefix)

            print("Using Carla version %s" % VERSION)
            print("  Python version: %s" % sys.version.split(" ",1)[0])
            print("  Qt version:     %s" % QT_VERSION_STR)
            print("  PyQt version:   %s" % PYQT_VERSION_STR)
            print("  Binary dir:     %s" % pathBinaries)
            print("  Resources dir:  %s" % pathResources)

            sys.exit(1 if gCarla.nogui else 0)

        elif readPrefixNext:
            readPrefixNext = False
            gCarla.cnprefix = arg

    if gCarla.nogui and not isinstance(gCarla.nogui, bool):
        if os.fork():
            # pylint: disable=protected-access
            os._exit(0)
            # pylint: enable=protected-access
        else:
            os.setsid()

    return (initName, libPrefix)

# ------------------------------------------------------------------------------------------------------------
# Get initial project file (as passed in the command-line parameters)

def getInitialProjectFile(skipExistCheck = False):
    # NOTE: PyQt mishandles unicode characters, we directly use sys.argv instead of qApp->arguments()
    # see https://riverbankcomputing.com/pipermail/pyqt/2015-January/035395.html
    args = sys.argv[1:]
    readPrefixNext = False
    for arg in args:
        if readPrefixNext:
            readPrefixNext = False
            continue
        if arg.startswith("--cnprefix="):
            continue
        if arg.startswith("--osc-gui="):
            continue
        if arg.startswith("--with-appname="):
            continue
        if arg.startswith("--with-libprefix="):
            continue
        if arg == "--cnprefix":
            readPrefixNext = True
            continue
        if arg in ("-n", "--n", "-no-gui", "--no-gui", "-nogui", "--nogui", "--gdb"):
            continue
        if MACOS and arg.startswith("-psn_"):
            continue
        arg = os.path.expanduser(arg)
        if skipExistCheck or os.path.exists(arg):
            return arg

    return None

# ------------------------------------------------------------------------------------------------------------
# Get paths (binaries, resources)

def getPaths(libPrefix = None):
    CWDl = CWD.lower()

    # adjust for special distros
    libdir  = os.path.basename(os.path.normpath(X_LIBDIR_X))  if X_LIBDIR_X  else "lib"
    datadir = os.path.basename(os.path.normpath(X_DATADIR_X)) if X_DATADIR_X else "share"

    # standalone, installed system-wide linux
    if libPrefix is not None:
        pathBinaries  = os.path.join(libPrefix, libdir, "carla")
        pathResources = os.path.join(libPrefix, datadir, "carla", "resources")

    # standalone, local source
    elif CWDl.endswith("frontend"):
        pathBinaries  = os.path.abspath(os.path.join(CWD, "..", "..", "bin"))
        pathResources = os.path.join(pathBinaries, "resources")

    # plugin
    elif CWDl.endswith("resources"):
        # installed system-wide linux
        if CWDl.endswith("/share/carla/resources"):
            pathBinaries  = os.path.abspath(os.path.join(CWD, "..", "..", "..", libdir, "carla"))
            pathResources = CWD

        # local source
        elif CWDl.endswith("native-plugins%sresources" % os.sep):
            pathBinaries  = os.path.abspath(os.path.join(CWD, "..", "..", "..", "bin"))
            pathResources = CWD

        # other
        else:
            pathBinaries  = os.path.abspath(os.path.join(CWD, ".."))
            pathResources = CWD

    # everything else
    else:
        pathBinaries  = CWD
        pathResources = os.path.join(pathBinaries, "resources")

    return (pathBinaries, pathResources)

# ------------------------------------------------------------------------------------------------------------
# Signal handler
# TODO move to carla_host.py or something

def signalHandler(sig, frame):
    if sig in (SIGINT, SIGTERM):
        gCarla.term = True
        if gCarla.gui is not None:
            gCarla.gui.SIGTERM.emit()

    elif haveSIGUSR1 and sig == SIGUSR1:
        if gCarla.gui is not None:
            gCarla.gui.SIGUSR1.emit()

def setUpSignals():
    signal(SIGINT,  signalHandler)
    signal(SIGTERM, signalHandler)

    if not haveSIGUSR1:
        return

    signal(SIGUSR1, signalHandler)

# ------------------------------------------------------------------------------------------------------------
# QLineEdit and QPushButton combo

def getAndSetPath(parent, lineEdit):
    newPath = QFileDialog.getExistingDirectory(parent, parent.tr("Set Path"), lineEdit.text(), QFileDialog.ShowDirsOnly)
    if newPath:
        lineEdit.setText(newPath)
    return newPath

# ------------------------------------------------------------------------------------------------------------
# Backwards-compatible horizontalAdvance/width call, depending on Qt version

def fontMetricsHorizontalAdvance(fontMetrics, string):
    if QT_VERSION >= 0x50b00:
        return fontMetrics.horizontalAdvance(string)
    return fontMetrics.width(string)

# ------------------------------------------------------------------------------------------------------------
# Custom QMessageBox which resizes itself to fit text

class QMessageBoxWithBetterWidth(QMessageBox):
    def __init__(self, parent):
        QMessageBox.__init__(self, parent)

    def showEvent(self, event):
        fontMetrics = self.fontMetrics()

        lines = self.text().strip().split("\n") + self.informativeText().strip().split("\n")

        if lines:
            width = 0

            for line in lines:
                width = max(fontMetricsHorizontalAdvance(fontMetrics, line), width)

            self.layout().setColumnMinimumWidth(2, width + 12)

        QMessageBox.showEvent(self, event)

# ------------------------------------------------------------------------------------------------------------
# Safer QSettings class, which does not throw if type mismatches

class QSafeSettings(QSettings):
    def value(self, key, defaultValue, valueType):
        if not isinstance(defaultValue, valueType):
            print("QSafeSettings.value() - defaultValue type mismatch for key", key)

        try:
            return QSettings.value(self, key, defaultValue, valueType)
        except:
            return defaultValue

# ------------------------------------------------------------------------------------------------------------
# Custom MessageBox

# pylint: disable=too-many-arguments
def CustomMessageBox(parent, icon, title, text,
                     extraText="",
                     buttons=QMessageBox.Yes|QMessageBox.No,
                     defButton=QMessageBox.No):
    msgBox = QMessageBoxWithBetterWidth(parent)
    msgBox.setIcon(icon)
    msgBox.setWindowTitle(title)
    msgBox.setText(text)
    msgBox.setInformativeText(extraText)
    msgBox.setStandardButtons(buttons)
    msgBox.setDefaultButton(defButton)
    return msgBox.exec_()
# pylint: enable=too-many-arguments

# ------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '../../resources/ui/carla_plugin_calf.ui'
#
# Created by: PyQt5 UI code generator 5.15.11
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_PluginWidget(object):
    def setupUi(self, PluginWidget):
        PluginWidget.setObjectName("PluginWidget")
        PluginWidget.resize(622, 90)
        PluginWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.verticalLayout = QtWidgets.QVBoxLayout(PluginWidget)
        self.verticalLayout.setContentsMargins(4, 3, 6, 3)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.layout_top = QtWidgets.QHBoxLayout()
        self.layout_top.setSpacing(1)
        self.layout_top.setObjectName("layout_top")
        self.w_buttons = QtWidgets.QWidget(PluginWidget)
        self.w_buttons.setMinimumSize(QtCore.QSize(80, 0))
        self.w_buttons.setMaximumSize(QtCore.QSize(80, 16777215))
        self.w_buttons.setObjectName("w_buttons")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.w_buttons)
        self.horizontalLayout_5.setContentsMargins(12, 0, 0, 0)
        self.horizontalLayout_5.setSpacing(2)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.b_enable = ScalableButton(self.w_buttons)
        self.b_enable.setCheckable(True)
        self.b_enable.setObjectName("b_enable")
        self.horizontalLayout_5.addWidget(self.b_enable)
        self.label_active = QtWidgets.QLabel(self.w_buttons)
        self.label_active.setObjectName("label_active")
        self.horizontalLayout_5.addWidget(self.label_active)
        self.layout_top.addWidget(self.w_buttons)
        self.line_sep = QtWidgets.QFrame(PluginWidget)
        self.line_sep.setMinimumSize(QtCore.QSize(0, 20))
        self.line_sep.setMaximumSize(QtCore.QSize(16777215, 20))
        self.line_sep.setLineWidth(0)
        self.line_sep.setMidLineWidth(1)
        self.line_sep.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_sep.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_sep.setObjectName("line_sep")
        self.layout_top.addWidget(self.line_sep)
        spacerItem = QtWidgets.QSpacerItem(8, 1, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.layout_top.addItem(spacerItem)
        self.label_name = QtWidgets.QLabel(PluginWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_name.setFont(font)
        self.label_name.setAlignment(QtCore.Qt.AlignCenter)
        self.label_name.setObjectName("label_name")
        self.layout_top.addWidget(self.label_name)
        spacerItem1 = QtWidgets.QSpacerItem(20, 1, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.layout_top.addItem(spacerItem1)
        self.layout_midi = QtWidgets.QVBoxLayout()
        self.layout_midi.setContentsMargins(-1, -1, 4, -1)
        self.layout_midi.setSpacing(0)
        self.layout_midi.setObjectName("layout_midi")
        self.label_midi = QtWidgets.QLabel(PluginWidget)
        self.label_midi.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_midi.setObjectName("label_midi")
        self.layout_midi.addWidget(self.label_midi)
        self.led_midi = LEDButton(PluginWidget)
        self.led_midi.setMinimumSize(QtCore.QSize(25, 25))
        self.led_midi.setMaximumSize(QtCore.QSize(25, 25))
        self.led_midi.setText("")
        self.led_midi.setIconSize(QtCore.QSize(25, 25))
        self.led_midi.setCheckable(True)
        self.led_midi.setObjectName("led_midi")
        self.layout_midi.addWidget(self.led_midi)
        self.layout_top.addLayout(self.layout_midi)
        self.layout_peak_in = QtWidgets.QVBoxLayout()
        self.layout_peak_in.setSpacing(0)
        self.layout_peak_in.setObjectName("layout_peak_in")
        self.label_audio_in = QtWidgets.QLabel(PluginWidget)
        self.label_audio_in.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_audio_in.setObjectName("label_audio_in")
        self.layout_peak_in.addWidget(self.label_audio_in)
        self.peak_in = DigitalPeakMeter(PluginWidget)
        self.peak_in.setMinimumSize(QtCore.QSize(150, 0))
        self.peak_in.setObjectName("peak_in")
        self.layout_peak_in.addWidget(self.peak_in)
        self.layout_top.addLayout(self.layout_peak_in)
        self.layout_peak_out = QtWidgets.QVBoxLayout()
        self.layout_peak_out.setSpacing(0)
        self.layout_peak_out.setObjectName("layout_peak_out")
        self.label_audio_out = QtWidgets.QLabel(PluginWidget)
        self.label_audio_out.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_audio_out.setObjectName("label_audio_out")
        self.layout_peak_out.addWidget(self.label_audio_out)
        self.peak_out = DigitalPeakMeter(PluginWidget)
        self.peak_out.setMinimumSize(QtCore.QSize(150, 0))
        self.peak_out.setObjectName("peak_out")
        self.layout_peak_out.addWidget(self.peak_out)
        self.layout_top.addLayout(self.layout_peak_out)
        self.verticalLayout.addLayout(self.layout_top)
        self.line = QtWidgets.QFrame(PluginWidget)
        self.line.setLineWidth(0)
        self.line.setMidLineWidth(1)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.layout_bottom = QtWidgets.QHBoxLayout()
        self.layout_bottom.setContentsMargins(6, 4, 4, -1)
        self.layout_bottom.setSpacing(1)
        self.layout_bottom.setObjectName("layout_bottom")
        self.w_buttons2 = QtWidgets.QHBoxLayout()
        self.w_buttons2.setSpacing(3)
        self.w_buttons2.setObjectName("w_buttons2")
        self.b_gui = ScalableButton(PluginWidget)
        self.b_gui.setCheckable(True)
        self.b_gui.setObjectName("b_gui")
        self.w_buttons2.addWidget(self.b_gui)
        self.b_edit = ScalableButton(PluginWidget)
        self.b_edit.setCheckable(True)
        self.b_edit.setObjectName("b_edit")
        self.w_buttons2.addWidget(self.b_edit)
        self.b_remove = ScalableButton(PluginWidget)
        self.b_remove.setObjectName("b_remove")
        self.w_buttons2.addWidget(self.b_remove)
        self.layout_bottom.addLayout(self.w_buttons2)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.layout_bottom.addItem(spacerItem2)
        self.w_knobs = QtWidgets.QWidget(PluginWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.w_knobs.sizePolicy().hasHeightForWidth())
        self.w_knobs.setSizePolicy(sizePolicy)
        self.w_knobs.setObjectName("w_knobs")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.w_knobs)
        self.horizontalLayout_4.setContentsMargins(4, 0, 4, 0)
        self.horizontalLayout_4.setSpacing(2)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.layout_bottom.addWidget(self.w_knobs)
        self.label_logo = QtWidgets.QLabel(PluginWidget)
        self.label_logo.setMinimumSize(QtCore.QSize(71, 30))
        self.label_logo.setMaximumSize(QtCore.QSize(71, 30))
        self.label_logo.setText("")
        self.label_logo.setPixmap(QtGui.QPixmap(":/bitmaps/logo_calf.png"))
        self.label_logo.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_logo.setObjectName("label_logo")
        self.layout_bottom.addWidget(self.label_logo)
        self.verticalLayout.addLayout(self.layout_bottom)

        self.retranslateUi(PluginWidget)
        QtCore.QMetaObject.connectSlotsByName(PluginWidget)

    def retranslateUi(self, PluginWidget):
        _translate = QtCore.QCoreApplication.translate
        PluginWidget.setWindowTitle(_translate("PluginWidget", "Frame"))
        self.b_enable.setText(_translate("PluginWidget", "Enable"))
        self.label_active.setText(_translate("PluginWidget", "On/Off"))
        self.label_name.setText(_translate("PluginWidget", "PluginName"))
        self.label_midi.setText(_translate("PluginWidget", "MIDI"))
        self.label_audio_in.setText(_translate("PluginWidget", "AUDIO IN"))
        self.label_audio_out.setText(_translate("PluginWidget", "AUDIO OUT"))
        self.b_gui.setText(_translate("PluginWidget", "GUI"))
        self.b_edit.setText(_translate("PluginWidget", "Edit"))
        self.b_remove.setText(_translate("PluginWidget", "Remove"))
from widgets.digitalpeakmeter import DigitalPeakMeter
from widgets.ledbutton import LEDButton
from widgets.scalablebutton import ScalableButton
import resources_rc



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Carla settings code
# Copyright (C) 2011-2022 Filipe Coelho <falktx@falktx.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# For a full copy of the GNU General Public License see the doc/GPL.txt file.

# ---------------------------------------------------------------------------------------------------------------------
# Imports (PyQt5)

from PyQt5.QtCore import pyqtSlot, QT_VERSION, Qt
from PyQt5.QtWidgets import QDialog, QDialogButtonBox, QFileDialog

# ---------------------------------------------------------------------------------------------------------------------
# Imports (Custom)

import ui_carla_settings
import ui_carla_settings_driver

from carla_backend import (
    LINUX, MACOS, WINDOWS,
    ENGINE_DRIVER_DEVICE_HAS_CONTROL_PANEL,
    ENGINE_DRIVER_DEVICE_CAN_TRIPLE_BUFFER,
    ENGINE_DRIVER_DEVICE_VARIABLE_BUFFER_SIZE,
    ENGINE_DRIVER_DEVICE_VARIABLE_SAMPLE_RATE,
    ENGINE_PROCESS_MODE_CONTINUOUS_RACK,
    ENGINE_OPTION_FILE_PATH,
    ENGINE_OPTION_PLUGIN_PATH,
    FILE_AUDIO,
    FILE_MIDI,
    PLUGIN_LADSPA,
    PLUGIN_DSSI,
    PLUGIN_LV2,
    PLUGIN_VST2,
    PLUGIN_VST3,
    PLUGIN_SF2,
    PLUGIN_SFZ,
    PLUGIN_JSFX
)

from carla_shared import (
    CARLA_KEY_MAIN_PROJECT_FOLDER,
    CARLA_KEY_MAIN_USE_PRO_THEME,
    CARLA_KEY_MAIN_PRO_THEME_COLOR,
    CARLA_KEY_MAIN_REFRESH_INTERVAL,
    CARLA_KEY_MAIN_CONFIRM_EXIT,
    CARLA_KEY_MAIN_CLASSIC_SKIN,
    CARLA_KEY_MAIN_SHOW_LOGS,
    CARLA_KEY_MAIN_SYSTEM_ICONS,
    CARLA_KEY_MAIN_EXPERIMENTAL,
    CARLA_KEY_CANVAS_THEME,
    CARLA_KEY_CANVAS_SIZE,
    CARLA_KEY_CANVAS_USE_BEZIER_LINES,
    CARLA_KEY_CANVAS_AUTO_HIDE_GROUPS,
    CARLA_KEY_CANVAS_AUTO_SELECT_ITEMS,
    CARLA_KEY_CANVAS_EYE_CANDY,
    CARLA_KEY_CANVAS_FANCY_EYE_CANDY,
    CARLA_KEY_CANVAS_USE_OPENGL,
    CARLA_KEY_CANVAS_ANTIALIASING,
    CARLA_KEY_CANVAS_HQ_ANTIALIASING,
    CARLA_KEY_CANVAS_INLINE_DISPLAYS,
    CARLA_KEY_CANVAS_FULL_REPAINTS,
    CARLA_KEY_ENGINE_DRIVER_PREFIX,
    CARLA_KEY_ENGINE_AUDIO_DRIVER,
    CARLA_KEY_ENGINE_PROCESS_MODE,
    CARLA_KEY_ENGINE_FORCE_STEREO,
    CARLA_KEY_ENGINE_PREFER_PLUGIN_BRIDGES,
    CARLA_KEY_ENGINE_PREFER_UI_BRIDGES,
    CARLA_KEY_ENGINE_MANAGE_UIS,
    CARLA_KEY_ENGINE_UIS_ALWAYS_ON_TOP,
    CARLA_KEY_ENGINE_MAX_PARAMETERS,
    CARLA_KEY_ENGINE_RESET_XRUNS,
    CARLA_KEY_ENGINE_UI_BRIDGES_TIMEOUT,
    CARLA_KEY_OSC_ENABLED,
    CARLA_KEY_OSC_TCP_PORT_ENABLED,
    CARLA_KEY_OSC_TCP_PORT_NUMBER,
    CARLA_KEY_OSC_TCP_PORT_RANDOM,
    CARLA_KEY_OSC_UDP_PORT_ENABLED,
    CARLA_KEY_OSC_UDP_PORT_NUMBER,
    CARLA_KEY_OSC_UDP_PORT_RANDOM,
    CARLA_KEY_PATHS_AUDIO,
    CARLA_KEY_PATHS_MIDI,
    CARLA_KEY_PATHS_LADSPA,
    CARLA_KEY_PATHS_DSSI,
    CARLA_KEY_PATHS_LV2,
    CARLA_KEY_PATHS_VST2,
    CARLA_KEY_PATHS_VST3,
    CARLA_KEY_PATHS_SF2,
    CARLA_KEY_PATHS_SFZ,
    CARLA_KEY_PATHS_JSFX,
    CARLA_KEY_WINE_EXECUTABLE,
    CARLA_KEY_WINE_AUTO_PREFIX,
    CARLA_KEY_WINE_FALLBACK_PREFIX,
    CARLA_KEY_WINE_RT_PRIO_ENABLED,
    CARLA_KEY_WINE_BASE_RT_PRIO,
    CARLA_KEY_WINE_SERVER_RT_PRIO,
    CARLA_KEY_EXPERIMENTAL_PLUGIN_BRIDGES,
    CARLA_KEY_EXPERIMENTAL_WINE_BRIDGES,
    CARLA_KEY_EXPERIMENTAL_JACK_APPS,
    CARLA_KEY_EXPERIMENTAL_EXPORT_LV2,
    CARLA_KEY_EXPERIMENTAL_PREVENT_BAD_BEHAVIOUR,
    CARLA_KEY_EXPERIMENTAL_LOAD_LIB_GLOBAL,
    CARLA_DEFAULT_MAIN_PROJECT_FOLDER,
    CARLA_DEFAULT_MAIN_USE_PRO_THEME,
    CARLA_DEFAULT_MAIN_PRO_THEME_COLOR,
    CARLA_DEFAULT_MAIN_REFRESH_INTERVAL,
    CARLA_DEFAULT_MAIN_CONFIRM_EXIT,
    CARLA_DEFAULT_MAIN_CLASSIC_SKIN,
    CARLA_DEFAULT_MAIN_SHOW_LOGS,
    CARLA_DEFAULT_MAIN_SYSTEM_ICONS,
    #CARLA_DEFAULT_MAIN_EXPERIMENTAL,
    CARLA_DEFAULT_CANVAS_THEME,
    CARLA_DEFAULT_CANVAS_SIZE,
    CARLA_DEFAULT_CANVAS_USE_BEZIER_LINES,
    CARLA_DEFAULT_CANVAS_AUTO_HIDE_GROUPS,
    CARLA_DEFAULT_CANVAS_AUTO_SELECT_ITEMS,
    CARLA_DEFAULT_CANVAS_EYE_CANDY,
    CARLA_DEFAULT_CANVAS_FANCY_EYE_CANDY,
    CARLA_DEFAULT_CANVAS_USE_OPENGL,
    CARLA_DEFAULT_CANVAS_ANTIALIASING,
    CARLA_DEFAULT_CANVAS_HQ_ANTIALIASING,
    CARLA_DEFAULT_CANVAS_INLINE_DISPLAYS,
    CARLA_DEFAULT_CANVAS_FULL_REPAINTS,
    CARLA_DEFAULT_FORCE_STEREO,
    CARLA_DEFAULT_PREFER_PLUGIN_BRIDGES,
    CARLA_DEFAULT_PREFER_UI_BRIDGES,
    CARLA_DEFAULT_MANAGE_UIS,
    CARLA_DEFAULT_UIS_ALWAYS_ON_TOP,
    CARLA_DEFAULT_MAX_PARAMETERS,
    CARLA_DEFAULT_RESET_XRUNS,
    CARLA_DEFAULT_UI_BRIDGES_TIMEOUT,
    CARLA_DEFAULT_AUDIO_BUFFER_SIZE,
    CARLA_DEFAULT_AUDIO_SAMPLE_RATE,
    CARLA_DEFAULT_AUDIO_TRIPLE_BUFFER,
    CARLA_DEFAULT_AUDIO_DRIVER,
    CARLA_DEFAULT_PROCESS_MODE,
    CARLA_DEFAULT_OSC_ENABLED,
    CARLA_DEFAULT_OSC_TCP_PORT_ENABLED,
    CARLA_DEFAULT_OSC_TCP_PORT_NUMBER,
    CARLA_DEFAULT_OSC_TCP_PORT_RANDOM,
    CARLA_DEFAULT_OSC_UDP_PORT_ENABLED,
    CARLA_DEFAULT_OSC_UDP_PORT_NUMBER,
    CARLA_DEFAULT_OSC_UDP_PORT_RANDOM,
    CARLA_DEFAULT_WINE_EXECUTABLE,
    CARLA_DEFAULT_WINE_AUTO_PREFIX,
    CARLA_DEFAULT_WINE_FALLBACK_PREFIX,
    CARLA_DEFAULT_WINE_RT_PRIO_ENABLED,
    CARLA_DEFAULT_WINE_BASE_RT_PRIO,
    CARLA_DEFAULT_WINE_SERVER_RT_PRIO,
    CARLA_DEFAULT_EXPERIMENTAL_PLUGIN_BRIDGES,
    CARLA_DEFAULT_EXPERIMENTAL_WINE_BRIDGES,
    CARLA_DEFAULT_EXPERIMENTAL_JACK_APPS,
    CARLA_DEFAULT_EXPERIMENTAL_LV2_EXPORT,
    CARLA_DEFAULT_EXPERIMENTAL_PREVENT_BAD_BEHAVIOUR,
    CARLA_DEFAULT_EXPERIMENTAL_LOAD_LIB_GLOBAL,
    CARLA_DEFAULT_FILE_PATH_AUDIO,
    CARLA_DEFAULT_FILE_PATH_MIDI,
    CARLA_DEFAULT_LADSPA_PATH,
    CARLA_DEFAULT_DSSI_PATH,
    CARLA_DEFAULT_LV2_PATH,
    CARLA_DEFAULT_VST2_PATH,
    CARLA_DEFAULT_VST3_PATH,
    CARLA_DEFAULT_SF2_PATH,
    CARLA_DEFAULT_SFZ_PATH,
    CARLA_DEFAULT_JSFX_PATH,
    getAndSetPath,
    getIcon,
    fontMetricsHorizontalAdvance,
    splitter,
    QSafeSettings
)

from patchcanvas.theme import Theme, getThemeName

# ---------------------------------------------------------------------------------------------------------------------
# ...

BUFFER_SIZE_LIST = (16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192)
SAMPLE_RATE_LIST = (22050, 32000, 44100, 48000, 88200, 96000, 176400, 192000)

# ---------------------------------------------------------------------------------------------------------------------
# Driver Settings

class DriverSettingsW(QDialog):
    AUTOMATIC_OPTION = "(Auto)"

    def __init__(self, parent, host, driverIndex, driverName):
        QDialog.__init__(self, parent)
        self.host = host
        self.ui = ui_carla_settings_driver.Ui_DriverSettingsW()
        self.ui.setupUi(self)

        # -------------------------------------------------------------------------------------------------------------
        # Internal stuff

        self.fDriverIndex = driverIndex
        self.fDriverName  = driverName
        self.fDeviceNames = host.get_engine_driver_device_names(driverIndex)

        self.fBufferSizes = BUFFER_SIZE_LIST
        self.fSampleRates = SAMPLE_RATE_LIST

        # -------------------------------------------------------------------------------------------------------------
        # Set-up GUI

        for name in self.fDeviceNames:
            self.ui.cb_device.addItem(name)

        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        if MACOS:
            self.setWindowModality(Qt.WindowModal)

        # -------------------------------------------------------------------------------------------------------------
        # Load settings

        self.loadSettings()

        # -------------------------------------------------------------------------------------------------------------
        # Set-up connections

        self.accepted.connect(self.slot_saveSettings)
        self.ui.b_panel.clicked.connect(self.slot_showDevicePanel)
        self.ui.cb_device.currentIndexChanged.connect(self.slot_updateDeviceInfo)

    # -----------------------------------------------------------------------------------------------------------------

    def getValues(self):
        audioDevice = self.ui.cb_device.currentText()
        bufferSize  = self.ui.cb_buffersize.currentText()
        sampleRate  = self.ui.cb_samplerate.currentText()

        if bufferSize == self.AUTOMATIC_OPTION:
            bufferSize = "0"
        if sampleRate == self.AUTOMATIC_OPTION:
            sampleRate = "0"

        return (audioDevice, int(bufferSize), int(sampleRate))

    # -----------------------------------------------------------------------------------------------------------------

    def loadSettings(self):
        settings = QSafeSettings("falkTX", "Carla2")

        args = (CARLA_KEY_ENGINE_DRIVER_PREFIX, self.fDriverName)
        audioDevice       = settings.value("%s%s/Device"       % args, "",                                str)
        audioBufferSize   = settings.value("%s%s/BufferSize"   % args, CARLA_DEFAULT_AUDIO_BUFFER_SIZE,   int)
        audioSampleRate   = settings.value("%s%s/SampleRate"   % args, CARLA_DEFAULT_AUDIO_SAMPLE_RATE,   int)
        audioTripleBuffer = settings.value("%s%s/TripleBuffer" % args, CARLA_DEFAULT_AUDIO_TRIPLE_BUFFER, bool)

        if audioDevice and audioDevice in self.fDeviceNames:
            self.ui.cb_device.setCurrentIndex(self.fDeviceNames.index(audioDevice))
        else:
            self.ui.cb_device.setCurrentIndex(-1)

        # fill combo-boxes first
        self.slot_updateDeviceInfo()

        if audioBufferSize and audioBufferSize in self.fBufferSizes:
            self.ui.cb_buffersize.setCurrentIndex(self.fBufferSizes.index(audioBufferSize))
        elif self.fBufferSizes == BUFFER_SIZE_LIST:
            self.ui.cb_buffersize.setCurrentIndex(BUFFER_SIZE_LIST.index(CARLA_DEFAULT_AUDIO_BUFFER_SIZE))
        else:
            self.ui.cb_buffersize.setCurrentIndex(int(len(self.fBufferSizes)/2))

        if audioSampleRate and audioSampleRate in self.fSampleRates:
            self.ui.cb_samplerate.setCurrentIndex(self.fSampleRates.index(audioSampleRate))
        elif self.fSampleRates == SAMPLE_RATE_LIST:
            self.ui.cb_samplerate.setCurrentIndex(SAMPLE_RATE_LIST.index(CARLA_DEFAULT_AUDIO_SAMPLE_RATE))
        else:
            self.ui.cb_samplerate.setCurrentIndex(int(len(self.fSampleRates)/2))

        self.ui.cb_triple_buffer.setChecked(audioTripleBuffer and self.ui.cb_triple_buffer.isEnabled())

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot()
    def slot_saveSettings(self):
        settings = QSafeSettings("falkTX", "Carla2")

        bufferSize = self.ui.cb_buffersize.currentText()
        sampleRate = self.ui.cb_samplerate.currentText()

        if bufferSize == self.AUTOMATIC_OPTION:
            bufferSize = "0"
        if sampleRate == self.AUTOMATIC_OPTION:
            sampleRate = "0"

        args = (CARLA_KEY_ENGINE_DRIVER_PREFIX, self.fDriverName)
        settings.setValue("%s%s/Device"       % args, self.ui.cb_device.currentText())
        settings.setValue("%s%s/BufferSize"   % args, bufferSize)
        settings.setValue("%s%s/SampleRate"   % args, sampleRate)
        settings.setValue("%s%s/TripleBuffer" % args, self.ui.cb_triple_buffer.isChecked())

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot()
    def slot_showDevicePanel(self):
        self.host.show_engine_driver_device_control_panel(self.fDriverIndex, self.ui.cb_device.currentText())

    @pyqtSlot()
    def slot_updateDeviceInfo(self):
        deviceName = self.ui.cb_device.currentText()

        oldBufferSize = self.ui.cb_buffersize.currentText()
        oldSampleRate = self.ui.cb_samplerate.currentText()

        self.ui.cb_buffersize.clear()
        self.ui.cb_samplerate.clear()

        driverDeviceInfo  = self.host.get_engine_driver_device_info(self.fDriverIndex, deviceName)
        driverDeviceHints = driverDeviceInfo['hints']
        self.fBufferSizes = driverDeviceInfo['bufferSizes']
        self.fSampleRates = driverDeviceInfo['sampleRates']

        if driverDeviceHints & ENGINE_DRIVER_DEVICE_CAN_TRIPLE_BUFFER:
            self.ui.cb_triple_buffer.setEnabled(True)
        else:
            self.ui.cb_triple_buffer.setEnabled(False)

        if driverDeviceHints & ENGINE_DRIVER_DEVICE_HAS_CONTROL_PANEL:
            self.ui.b_panel.setEnabled(True)
        else:
            self.ui.b_panel.setEnabled(False)

        if self.fBufferSizes:
            match = False
            for bsize in self.fBufferSizes:
                sbsize = str(bsize)
                self.ui.cb_buffersize.addItem(sbsize)

                if oldBufferSize == sbsize:
                    match = True
                    self.ui.cb_buffersize.setCurrentIndex(self.ui.cb_buffersize.count()-1)

            if not match:
                if CARLA_DEFAULT_AUDIO_BUFFER_SIZE in self.fBufferSizes:
                    self.ui.cb_buffersize.setCurrentIndex(self.fBufferSizes.index(CARLA_DEFAULT_AUDIO_BUFFER_SIZE))
                else:
                    self.ui.cb_buffersize.setCurrentIndex(int(len(self.fBufferSizes)/2))

        else:
            self.ui.cb_buffersize.addItem(self.AUTOMATIC_OPTION)
            self.ui.cb_buffersize.setCurrentIndex(0)

        if self.fSampleRates:
            match = False
            for srate in self.fSampleRates:
                ssrate = str(int(srate))
                self.ui.cb_samplerate.addItem(ssrate)

                if oldSampleRate == ssrate:
                    match = True
                    self.ui.cb_samplerate.setCurrentIndex(self.ui.cb_samplerate.count()-1)

            if not match:
                if CARLA_DEFAULT_AUDIO_SAMPLE_RATE in self.fSampleRates:
                    self.ui.cb_samplerate.setCurrentIndex(self.fSampleRates.index(CARLA_DEFAULT_AUDIO_SAMPLE_RATE))
                else:
                    self.ui.cb_samplerate.setCurrentIndex(int(len(self.fSampleRates)/2))

        else:
            self.ui.cb_samplerate.addItem(self.AUTOMATIC_OPTION)
            self.ui.cb_samplerate.setCurrentIndex(0)

# ---------------------------------------------------------------------------------------------------------------------
# Runtime Driver Settings

class RuntimeDriverSettingsW(QDialog):
    def __init__(self, parent, host):
        QDialog.__init__(self, parent)
        self.host = host
        self.ui = ui_carla_settings_driver.Ui_DriverSettingsW()
        self.ui.setupUi(self)

        driverDeviceInfo = host.get_runtime_engine_driver_device_info()

        # -------------------------------------------------------------------------------------------------------------
        # Set-up GUI

        self.ui.cb_device.clear()
        self.ui.cb_buffersize.clear()
        self.ui.cb_samplerate.clear()
        self.ui.cb_triple_buffer.hide()
        self.ui.ico_restart.hide()
        self.ui.label_restart.hide()

        self.ui.layout_triple_buffer.takeAt(2)
        self.ui.layout_triple_buffer.takeAt(1)
        self.ui.layout_triple_buffer.takeAt(0)
        self.ui.verticalLayout.removeItem(self.ui.layout_triple_buffer)

        self.ui.layout_restart.takeAt(3)
        self.ui.layout_restart.takeAt(2)
        self.ui.layout_restart.takeAt(1)
        self.ui.layout_restart.takeAt(0)
        self.ui.verticalLayout.removeItem(self.ui.layout_restart)

        self.adjustSize()
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        if MACOS:
            self.setWindowModality(Qt.WindowModal)

        # -------------------------------------------------------------------------------------------------------------
        # Load runtime settings

        if host.is_engine_running():
            self.ui.cb_device.addItem(driverDeviceInfo['name'])
            self.ui.cb_device.setCurrentIndex(0)
            self.ui.cb_device.setEnabled(False)
        else:
            self.ui.cb_device.addItem(driverDeviceInfo['name'])
            self.ui.cb_device.setCurrentIndex(0)

        if driverDeviceInfo['bufferSizes']:
            for bsize in driverDeviceInfo['bufferSizes']:
                sbsize = str(bsize)
                self.ui.cb_buffersize.addItem(sbsize)

                if driverDeviceInfo['bufferSize'] == bsize:
                    self.ui.cb_buffersize.setCurrentIndex(self.ui.cb_buffersize.count()-1)

        else:
            self.ui.cb_buffersize.addItem(DriverSettingsW.AUTOMATIC_OPTION)
            self.ui.cb_buffersize.setCurrentIndex(0)

        if (driverDeviceInfo['hints'] & ENGINE_DRIVER_DEVICE_VARIABLE_BUFFER_SIZE) == 0x0:
            self.ui.cb_buffersize.setEnabled(False)

        if driverDeviceInfo['sampleRates']:
            for srate in driverDeviceInfo['sampleRates']:
                ssrate = str(int(srate))
                self.ui.cb_samplerate.addItem(ssrate)

                if driverDeviceInfo['sampleRate'] == srate:
                    self.ui.cb_samplerate.setCurrentIndex(self.ui.cb_samplerate.count()-1)

        else:
            self.ui.cb_samplerate.addItem(DriverSettingsW.AUTOMATIC_OPTION)
            self.ui.cb_samplerate.setCurrentIndex(0)

        if (driverDeviceInfo['hints'] & ENGINE_DRIVER_DEVICE_VARIABLE_SAMPLE_RATE) == 0x0:
            self.ui.cb_samplerate.setEnabled(False)

        if (driverDeviceInfo['hints'] & ENGINE_DRIVER_DEVICE_HAS_CONTROL_PANEL) == 0x0:
            self.ui.b_panel.setEnabled(False)

        # -------------------------------------------------------------------------------------------------------------
        # Set-up connections

        self.ui.b_panel.clicked.connect(self.slot_showDevicePanel)

    # -----------------------------------------------------------------------------------------------------------------

    def getValues(self):
        audioDevice = self.ui.cb_device.currentText()
        bufferSize  = self.ui.cb_buffersize.currentText()
        sampleRate  = self.ui.cb_samplerate.currentText()

        if bufferSize == DriverSettingsW.AUTOMATIC_OPTION:
            bufferSize = "0"
        if sampleRate == DriverSettingsW.AUTOMATIC_OPTION:
            sampleRate = "0"

        return (audioDevice, int(bufferSize), int(sampleRate))

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot()
    def slot_showDevicePanel(self):
        self.host.show_engine_device_control_panel()

# ---------------------------------------------------------------------------------------------------------------------
# Settings Dialog

class CarlaSettingsW(QDialog):
    # Tab indexes
    TAB_INDEX_MAIN         = 0
    TAB_INDEX_CANVAS       = 1
    TAB_INDEX_ENGINE       = 2
    TAB_INDEX_OSC          = 3
    TAB_INDEX_FILEPATHS    = 4
    TAB_INDEX_PLUGINPATHS  = 5
    TAB_INDEX_WINE         = 6
    TAB_INDEX_EXPERIMENTAL = 7
    TAB_INDEX_NONE         = 8

    # File Path indexes
    FILEPATH_INDEX_AUDIO = 0
    FILEPATH_INDEX_MIDI  = 1

    # Plugin Path indexes
    PLUGINPATH_INDEX_LADSPA = 0
    PLUGINPATH_INDEX_DSSI   = 1
    PLUGINPATH_INDEX_LV2    = 2
    PLUGINPATH_INDEX_VST2   = 3
    PLUGINPATH_INDEX_VST3   = 4
    PLUGINPATH_INDEX_SF2    = 5
    PLUGINPATH_INDEX_SFZ    = 6
    PLUGINPATH_INDEX_JSFX   = 7

    # Single and Multiple client mode is only for JACK,
    # but we still want to match QComboBox index to backend defines,
    # so add +2 pos padding if driverName != "JACK".
    PROCESS_MODE_NON_JACK_PADDING = 2

    # -----------------------------------------------------------------------------------------------------------------

    def __init__(self, parent, host, hasCanvas, hasCanvasGL):
        QDialog.__init__(self, parent)
        self.host = host
        self.ui = ui_carla_settings.Ui_CarlaSettingsW()
        self.ui.setupUi(self)

        # -------------------------------------------------------------------------------------------------------------
        # Set-up GUI

        self.ui.lw_page.setFixedWidth(48 + 6*3 +
                                      fontMetricsHorizontalAdvance(self.ui.lw_page.fontMetrics(), "  Experimental  "))

        for i in range(host.get_engine_driver_count()):
            self.ui.cb_engine_audio_driver.addItem(host.get_engine_driver_name(i))

        for i in range(Theme.THEME_MAX):
            self.ui.cb_canvas_theme.addItem(getThemeName(i))

        if MACOS:
            self.ui.group_main_theme.setEnabled(False)
            self.ui.group_main_theme.setVisible(False)

        if WINDOWS or host.isControl:
            self.ui.ch_main_show_logs.setEnabled(False)
            self.ui.ch_main_show_logs.setVisible(False)

        if host.isControl:
            self.ui.lw_page.hideRow(self.TAB_INDEX_ENGINE)
            self.ui.lw_page.hideRow(self.TAB_INDEX_FILEPATHS)
            self.ui.lw_page.hideRow(self.TAB_INDEX_PLUGINPATHS)
            self.ui.ch_exp_export_lv2.hide()
            self.ui.group_experimental_engine.hide()
            self.ui.cb_canvas_inline_displays.hide()

        elif not hasCanvas:
            self.ui.lw_page.hideRow(self.TAB_INDEX_CANVAS)

        elif not hasCanvasGL:
            self.ui.cb_canvas_use_opengl.setEnabled(False)
            self.ui.cb_canvas_render_hq_aa.setEnabled(False)

        if host.isPlugin:
            self.ui.cb_engine_audio_driver.setEnabled(False)

        if host.audioDriverForced is not None:
            self.ui.cb_engine_audio_driver.setEnabled(False)
            self.ui.tb_engine_driver_config.setEnabled(False)

        if host.processModeForced:
            self.ui.cb_engine_process_mode_jack.setEnabled(False)
            self.ui.cb_engine_process_mode_other.setEnabled(False)

            if self.host.processMode == ENGINE_PROCESS_MODE_CONTINUOUS_RACK:
                self.ui.ch_engine_force_stereo.setEnabled(False)

        if host.isControl or host.isPlugin:
            self.ui.ch_main_confirm_exit.hide()
            self.ui.ch_exp_load_lib_global.hide()
            self.ui.lw_page.hideRow(self.TAB_INDEX_OSC)
            self.ui.lw_page.hideRow(self.TAB_INDEX_WINE)

        if not LINUX:
            self.ui.ch_exp_wine_bridges.setVisible(False)
            self.ui.ch_exp_jack_apps.setVisible(False)
            self.ui.ch_exp_prevent_bad_behaviour.setVisible(False)
            self.ui.lw_page.hideRow(self.TAB_INDEX_WINE)

        if not MACOS:
            self.ui.label_engine_ui_bridges_mac_note.setVisible(False)

        # FIXME, not implemented yet
        self.ui.ch_engine_uis_always_on_top.hide()

        # FIXME broken in some Qt5 versions
        self.ui.cb_canvas_eyecandy.setEnabled(QT_VERSION >= 0x50c00)

        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        if MACOS:
            self.setWindowModality(Qt.WindowModal)

        # -------------------------------------------------------------------------------------------------------------
        # Load settings

        self.loadSettings()

        # -------------------------------------------------------------------------------------------------------------
        # Set-up Icons

        if self.ui.ch_main_system_icons.isChecked():
            self.ui.b_main_proj_folder_open.setIcon(getIcon('document-open', 16, 'svgz'))
            self.ui.b_filepaths_add.setIcon(getIcon('list-add', 16, 'svgz'))
            self.ui.b_filepaths_change.setIcon(getIcon('edit-rename', 16, 'svgz'))
            self.ui.b_filepaths_remove.setIcon(getIcon('list-remove', 16, 'svgz'))
            self.ui.b_paths_add.setIcon(getIcon('list-add', 16, 'svgz'))
            self.ui.b_paths_change.setIcon(getIcon('edit-rename', 16, 'svgz'))
            self.ui.b_paths_remove.setIcon(getIcon('list-remove', 16, 'svgz'))

        # -------------------------------------------------------------------------------------------------------------
        # Set-up connections

        self.accepted.connect(self.slot_saveSettings)
        self.ui.buttonBox.button(QDialogButtonBox.Reset).clicked.connect(self.slot_resetSettings)

        self.ui.b_main_proj_folder_open.clicked.connect(self.slot_getAndSetProjectPath)

        self.ui.cb_engine_audio_driver.currentIndexChanged.connect(self.slot_engineAudioDriverChanged)
        self.ui.tb_engine_driver_config.clicked.connect(self.slot_showAudioDriverSettings)

        self.ui.b_paths_add.clicked.connect(self.slot_addPluginPath)
        self.ui.b_paths_remove.clicked.connect(self.slot_removePluginPath)
        self.ui.b_paths_change.clicked.connect(self.slot_changePluginPath)
        self.ui.cb_paths.currentIndexChanged.connect(self.slot_pluginPathTabChanged)
        self.ui.lw_ladspa.currentRowChanged.connect(self.slot_pluginPathRowChanged)
        self.ui.lw_dssi.currentRowChanged.connect(self.slot_pluginPathRowChanged)
        self.ui.lw_lv2.currentRowChanged.connect(self.slot_pluginPathRowChanged)
        self.ui.lw_vst.currentRowChanged.connect(self.slot_pluginPathRowChanged)
        self.ui.lw_vst3.currentRowChanged.connect(self.slot_pluginPathRowChanged)
        self.ui.lw_sf2.currentRowChanged.connect(self.slot_pluginPathRowChanged)
        self.ui.lw_sfz.currentRowChanged.connect(self.slot_pluginPathRowChanged)
        self.ui.lw_jsfx.currentRowChanged.connect(self.slot_pluginPathRowChanged)

        self.ui.b_filepaths_add.clicked.connect(self.slot_addFilePath)
        self.ui.b_filepaths_remove.clicked.connect(self.slot_removeFilePath)
        self.ui.b_filepaths_change.clicked.connect(self.slot_changeFilePath)
        self.ui.cb_filepaths.currentIndexChanged.connect(self.slot_filePathTabChanged)
        self.ui.lw_files_audio.currentRowChanged.connect(self.slot_filePathRowChanged)
        self.ui.lw_files_midi.currentRowChanged.connect(self.slot_filePathRowChanged)

        self.ui.ch_main_experimental.toggled.connect(self.slot_enableExperimental)
        self.ui.ch_exp_wine_bridges.toggled.connect(self.slot_enableWineBridges)
        self.ui.cb_exp_plugin_bridges.toggled.connect(self.slot_pluginBridgesToggled)
        self.ui.cb_canvas_eyecandy.toggled.connect(self.slot_canvasEyeCandyToggled)
        #self.ui.cb_canvas_fancy_eyecandy.toggled.connect(self.slot_canvasFancyEyeCandyToggled)
        self.ui.cb_canvas_use_opengl.toggled.connect(self.slot_canvasOpenGLToggled)

        # -------------------------------------------------------------------------------------------------------------
        # Post-connect setup

        self.ui.lw_ladspa.setCurrentRow(0)
        self.ui.lw_dssi.setCurrentRow(0)
        self.ui.lw_lv2.setCurrentRow(0)
        self.ui.lw_vst.setCurrentRow(0)
        self.ui.lw_vst3.setCurrentRow(0)
        self.ui.lw_sf2.setCurrentRow(0)
        self.ui.lw_sfz.setCurrentRow(0)
        self.ui.lw_jsfx.setCurrentRow(0)

        self.ui.lw_files_audio.setCurrentRow(0)
        self.ui.lw_files_midi.setCurrentRow(0)

        self.ui.lw_page.setCurrentCell(0, 0)

        self.slot_filePathTabChanged(0)
        self.slot_pluginPathTabChanged(0)

        self.adjustSize()

    # -----------------------------------------------------------------------------------------------------------------

    def loadSettings(self):
        settings = QSafeSettings()

        # -------------------------------------------------------------------------------------------------------------
        # Main

        self.ui.ch_main_show_logs.setChecked(self.host.showLogs)
        self.ui.ch_engine_uis_always_on_top.setChecked(self.host.uisAlwaysOnTop)

        self.ui.le_main_proj_folder.setText(
            settings.value(CARLA_KEY_MAIN_PROJECT_FOLDER, CARLA_DEFAULT_MAIN_PROJECT_FOLDER, str))

        self.ui.ch_main_theme_pro.setChecked(self.ui.group_main_theme.isEnabled() and
                                             settings.value(CARLA_KEY_MAIN_USE_PRO_THEME,
                                                            CARLA_DEFAULT_MAIN_USE_PRO_THEME, bool))

        self.ui.cb_main_theme_color.setCurrentIndex(
            self.ui.cb_main_theme_color.findText(settings.value(CARLA_KEY_MAIN_PRO_THEME_COLOR,
                                                                CARLA_DEFAULT_MAIN_PRO_THEME_COLOR, str)))

        self.ui.sb_main_refresh_interval.setValue(
            settings.value(CARLA_KEY_MAIN_REFRESH_INTERVAL, CARLA_DEFAULT_MAIN_REFRESH_INTERVAL, int))

        self.ui.ch_main_confirm_exit.setChecked(
            settings.value(CARLA_KEY_MAIN_CONFIRM_EXIT, CARLA_DEFAULT_MAIN_CONFIRM_EXIT, bool))

        self.ui.cb_main_classic_skin_default.setChecked(
            settings.value(CARLA_KEY_MAIN_CLASSIC_SKIN, CARLA_DEFAULT_MAIN_CLASSIC_SKIN, bool))

        self.ui.ch_main_system_icons.setChecked(
            settings.value(CARLA_KEY_MAIN_SYSTEM_ICONS, CARLA_DEFAULT_MAIN_SYSTEM_ICONS, bool))

        # -------------------------------------------------------------------------------------------------------------
        # Canvas

        self.ui.cb_canvas_theme.setCurrentIndex(
            self.ui.cb_canvas_theme.findText(settings.value(CARLA_KEY_CANVAS_THEME, CARLA_DEFAULT_CANVAS_THEME, str)))

        self.ui.cb_canvas_size.setCurrentIndex(
            self.ui.cb_canvas_size.findText(settings.value(CARLA_KEY_CANVAS_SIZE, CARLA_DEFAULT_CANVAS_SIZE, str)))

        self.ui.cb_canvas_bezier_lines.setChecked(
            settings.value(CARLA_KEY_CANVAS_USE_BEZIER_LINES, CARLA_DEFAULT_CANVAS_USE_BEZIER_LINES, bool))

        self.ui.cb_canvas_hide_groups.setChecked(
            settings.value(CARLA_KEY_CANVAS_AUTO_HIDE_GROUPS, CARLA_DEFAULT_CANVAS_AUTO_HIDE_GROUPS, bool))

        self.ui.cb_canvas_auto_select.setChecked(
            settings.value(CARLA_KEY_CANVAS_AUTO_SELECT_ITEMS, CARLA_DEFAULT_CANVAS_AUTO_SELECT_ITEMS, bool))

        if self.ui.cb_canvas_eyecandy.isEnabled():
            self.ui.cb_canvas_eyecandy.setChecked(
                settings.value(CARLA_KEY_CANVAS_EYE_CANDY, CARLA_DEFAULT_CANVAS_EYE_CANDY, bool))

        self.ui.cb_canvas_fancy_eyecandy.setChecked(
            settings.value(CARLA_KEY_CANVAS_FANCY_EYE_CANDY, CARLA_DEFAULT_CANVAS_FANCY_EYE_CANDY, bool))

        self.ui.cb_canvas_use_opengl.setChecked(self.ui.cb_canvas_use_opengl.isEnabled() and
                                                settings.value(CARLA_KEY_CANVAS_USE_OPENGL,
                                                               CARLA_DEFAULT_CANVAS_USE_OPENGL, bool))

        self.ui.cb_canvas_render_aa.setCheckState(
            settings.value(CARLA_KEY_CANVAS_ANTIALIASING, CARLA_DEFAULT_CANVAS_ANTIALIASING, int))

        self.ui.cb_canvas_render_hq_aa.setChecked(self.ui.cb_canvas_render_hq_aa.isEnabled() and
                                                  settings.value(CARLA_KEY_CANVAS_HQ_ANTIALIASING,
                                                                 CARLA_DEFAULT_CANVAS_HQ_ANTIALIASING, bool))

        self.ui.cb_canvas_full_repaints.setChecked(
            settings.value(CARLA_KEY_CANVAS_FULL_REPAINTS, CARLA_DEFAULT_CANVAS_FULL_REPAINTS, bool))

        self.ui.cb_canvas_inline_displays.setChecked(
            settings.value(CARLA_KEY_CANVAS_INLINE_DISPLAYS, CARLA_DEFAULT_CANVAS_INLINE_DISPLAYS, bool))

        # -------------------------------------------------------------------------------------------------------------

        settings = QSafeSettings("falkTX", "Carla2")

        # -------------------------------------------------------------------------------------------------------------
        # Main

        self.ui.ch_main_experimental.setChecked(self.host.experimental)

        if not self.host.experimental:
            self.ui.lw_page.hideRow(self.TAB_INDEX_EXPERIMENTAL)
            self.ui.lw_page.hideRow(self.TAB_INDEX_WINE)

        elif not self.host.showWineBridges:
            self.ui.lw_page.hideRow(self.TAB_INDEX_WINE)

        # -------------------------------------------------------------------------------------------------------------
        # Engine

        if self.host.isPlugin:
            audioDriver = "Plugin"
            self.ui.cb_engine_audio_driver.setCurrentIndex(0)
        elif self.host.audioDriverForced:
            audioDriver = self.host.audioDriverForced
            self.ui.cb_engine_audio_driver.setCurrentIndex(0)
        else:
            audioDriver = settings.value(CARLA_KEY_ENGINE_AUDIO_DRIVER, CARLA_DEFAULT_AUDIO_DRIVER, str)

            for i in range(self.ui.cb_engine_audio_driver.count()):
                if self.ui.cb_engine_audio_driver.itemText(i) == audioDriver:
                    self.ui.cb_engine_audio_driver.setCurrentIndex(i)
                    break
            else:
                self.ui.cb_engine_audio_driver.setCurrentIndex(-1)

        if audioDriver == "JACK":
            self.ui.sw_engine_process_mode.setCurrentIndex(0)
        else:
            self.ui.sw_engine_process_mode.setCurrentIndex(1)

        self.ui.tb_engine_driver_config.setEnabled(self.host.audioDriverForced is None and not self.host.isPlugin)

        self.ui.cb_engine_process_mode_jack.setCurrentIndex(self.host.nextProcessMode)

        if self.host.nextProcessMode >= self.PROCESS_MODE_NON_JACK_PADDING:
            self.ui.cb_engine_process_mode_other.setCurrentIndex(self.host.nextProcessMode -
                                                                 self.PROCESS_MODE_NON_JACK_PADDING)
        else:
            self.ui.cb_engine_process_mode_other.setCurrentIndex(0)

        self.ui.sb_engine_max_params.setValue(self.host.maxParameters)
        self.ui.cb_engine_reset_xruns.setChecked(self.host.resetXruns)
        self.ui.ch_engine_manage_uis.setChecked(self.host.manageUIs)
        self.ui.ch_engine_prefer_ui_bridges.setChecked(self.host.preferUIBridges)
        self.ui.sb_engine_ui_bridges_timeout.setValue(self.host.uiBridgesTimeout)
        self.ui.ch_engine_force_stereo.setChecked(self.host.forceStereo or
                                                  not self.ui.ch_engine_force_stereo.isEnabled())
        self.ui.ch_engine_prefer_plugin_bridges.setChecked(self.host.preferPluginBridges)
        self.ui.ch_exp_export_lv2.setChecked(self.host.exportLV2)
        self.ui.cb_exp_plugin_bridges.setChecked(self.host.showPluginBridges)
        self.ui.ch_exp_wine_bridges.setChecked(self.host.showWineBridges)

        # -------------------------------------------------------------------------------------------------------------
        # OSC

        self.ui.ch_osc_enable.setChecked(
            settings.value(CARLA_KEY_OSC_ENABLED, CARLA_DEFAULT_OSC_ENABLED, bool))

        self.ui.group_osc_tcp_port.setChecked(
            settings.value(CARLA_KEY_OSC_TCP_PORT_ENABLED, CARLA_DEFAULT_OSC_TCP_PORT_ENABLED, bool))

        self.ui.group_osc_udp_port.setChecked(
            settings.value(CARLA_KEY_OSC_UDP_PORT_ENABLED, CARLA_DEFAULT_OSC_UDP_PORT_ENABLED, bool))

        self.ui.sb_osc_tcp_port_number.setValue(
            settings.value(CARLA_KEY_OSC_TCP_PORT_NUMBER, CARLA_DEFAULT_OSC_TCP_PORT_NUMBER, int))

        self.ui.sb_osc_udp_port_number.setValue(
            settings.value(CARLA_KEY_OSC_UDP_PORT_NUMBER, CARLA_DEFAULT_OSC_UDP_PORT_NUMBER, int))

        if settings.value(CARLA_KEY_OSC_TCP_PORT_RANDOM, CARLA_DEFAULT_OSC_TCP_PORT_RANDOM, bool):
            self.ui.rb_osc_tcp_port_specific.setChecked(False)
            self.ui.rb_osc_tcp_port_random.setChecked(True)
        else:
            self.ui.rb_osc_tcp_port_random.setChecked(False)
            self.ui.rb_osc_tcp_port_specific.setChecked(True)

        if settings.value(CARLA_KEY_OSC_UDP_PORT_RANDOM, CARLA_DEFAULT_OSC_UDP_PORT_RANDOM, bool):
            self.ui.rb_osc_udp_port_specific.setChecked(False)
            self.ui.rb_osc_udp_port_random.setChecked(True)
        else:
            self.ui.rb_osc_udp_port_random.setChecked(False)
            self.ui.rb_osc_udp_port_specific.setChecked(True)

        # -------------------------------------------------------------------------------------------------------------
        # File Paths

        audioPaths = settings.value(CARLA_KEY_PATHS_AUDIO, CARLA_DEFAULT_FILE_PATH_AUDIO, list)
        midiPaths = settings.value(CARLA_KEY_PATHS_MIDI, CARLA_DEFAULT_FILE_PATH_MIDI, list)

        audioPaths.sort()
        midiPaths.sort()

        for audioPath in audioPaths:
            if not audioPath:
                continue
            self.ui.lw_files_audio.addItem(audioPath)

        for midiPath in midiPaths:
            if not midiPath:
                continue
            self.ui.lw_files_midi.addItem(midiPath)

        # -------------------------------------------------------------------------------------------------------------
        # Plugin Paths

        ladspas = settings.value(CARLA_KEY_PATHS_LADSPA, CARLA_DEFAULT_LADSPA_PATH, list)
        dssis   = settings.value(CARLA_KEY_PATHS_DSSI,   CARLA_DEFAULT_DSSI_PATH, list)
        lv2s    = settings.value(CARLA_KEY_PATHS_LV2,    CARLA_DEFAULT_LV2_PATH, list)
        vst2s   = settings.value(CARLA_KEY_PATHS_VST2,   CARLA_DEFAULT_VST2_PATH, list)
        vst3s   = settings.value(CARLA_KEY_PATHS_VST3,   CARLA_DEFAULT_VST3_PATH, list)
        sf2s    = settings.value(CARLA_KEY_PATHS_SF2,    CARLA_DEFAULT_SF2_PATH, list)
        sfzs    = settings.value(CARLA_KEY_PATHS_SFZ,    CARLA_DEFAULT_SFZ_PATH, list)
        jsfxs   = settings.value(CARLA_KEY_PATHS_JSFX,   CARLA_DEFAULT_JSFX_PATH, list)

        ladspas.sort()
        dssis.sort()
        lv2s.sort()
        vst2s.sort()
        vst3s.sort()
        sf2s.sort()
        sfzs.sort()
        jsfxs.sort()

        for ladspa in ladspas:
            if not ladspa:
                continue
            self.ui.lw_ladspa.addItem(ladspa)

        for dssi in dssis:
            if not dssi:
                continue
            self.ui.lw_dssi.addItem(dssi)

        for lv2 in lv2s:
            if not lv2:
                continue
            self.ui.lw_lv2.addItem(lv2)

        for vst2 in vst2s:
            if not vst2:
                continue
            self.ui.lw_vst.addItem(vst2)

        for vst3 in vst3s:
            if not vst3:
                continue
            self.ui.lw_vst3.addItem(vst3)

        for sf2 in sf2s:
            if not sf2:
                continue
            self.ui.lw_sf2.addItem(sf2)

        for sfz in sfzs:
            if not sfz:
                continue
            self.ui.lw_sfz.addItem(sfz)

        for jsfx in jsfxs:
            if not jsfx:
                continue
            self.ui.lw_jsfx.addItem(jsfx)

        # -------------------------------------------------------------------------------------------------------------
        # Wine

        self.ui.le_wine_exec.setText(
            settings.value(CARLA_KEY_WINE_EXECUTABLE, CARLA_DEFAULT_WINE_EXECUTABLE, str))

        self.ui.cb_wine_prefix_detect.setChecked(
            settings.value(CARLA_KEY_WINE_AUTO_PREFIX, CARLA_DEFAULT_WINE_AUTO_PREFIX, bool))

        self.ui.le_wine_prefix_fallback.setText(
            settings.value(CARLA_KEY_WINE_FALLBACK_PREFIX, CARLA_DEFAULT_WINE_FALLBACK_PREFIX, str))

        self.ui.group_wine_realtime.setChecked(
            settings.value(CARLA_KEY_WINE_RT_PRIO_ENABLED, CARLA_DEFAULT_WINE_RT_PRIO_ENABLED, bool))

        self.ui.sb_wine_base_prio.setValue(
            settings.value(CARLA_KEY_WINE_BASE_RT_PRIO, CARLA_DEFAULT_WINE_BASE_RT_PRIO, int))

        self.ui.sb_wine_server_prio.setValue(
            settings.value(CARLA_KEY_WINE_SERVER_RT_PRIO, CARLA_DEFAULT_WINE_SERVER_RT_PRIO, int))

        # -------------------------------------------------------------------------------------------------------------
        # Experimental

        self.ui.ch_exp_jack_apps.setChecked(
            settings.value(CARLA_KEY_EXPERIMENTAL_JACK_APPS, CARLA_DEFAULT_EXPERIMENTAL_JACK_APPS, bool))

        self.ui.ch_exp_export_lv2.setChecked(
            settings.value(CARLA_KEY_EXPERIMENTAL_EXPORT_LV2, CARLA_DEFAULT_EXPERIMENTAL_LV2_EXPORT, bool))

        self.ui.ch_exp_load_lib_global.setChecked(
            settings.value(CARLA_KEY_EXPERIMENTAL_LOAD_LIB_GLOBAL, CARLA_DEFAULT_EXPERIMENTAL_LOAD_LIB_GLOBAL, bool))

        self.ui.ch_exp_prevent_bad_behaviour.setChecked(
            settings.value(CARLA_KEY_EXPERIMENTAL_PREVENT_BAD_BEHAVIOUR,
                           CARLA_DEFAULT_EXPERIMENTAL_PREVENT_BAD_BEHAVIOUR, bool))

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot()
    def slot_saveSettings(self):
        settings = QSafeSettings()

        self.host.experimental = self.ui.ch_main_experimental.isChecked()

        if not self.host.experimental:
            self.resetExperimentalSettings()

        # -------------------------------------------------------------------------------------------------------------
        # Main

        settings.setValue(CARLA_KEY_MAIN_PROJECT_FOLDER,   self.ui.le_main_proj_folder.text())
        settings.setValue(CARLA_KEY_MAIN_CONFIRM_EXIT,     self.ui.ch_main_confirm_exit.isChecked())
        settings.setValue(CARLA_KEY_MAIN_CLASSIC_SKIN,     self.ui.cb_main_classic_skin_default.isChecked())
        settings.setValue(CARLA_KEY_MAIN_USE_PRO_THEME,    self.ui.ch_main_theme_pro.isChecked())
        settings.setValue(CARLA_KEY_MAIN_PRO_THEME_COLOR,  self.ui.cb_main_theme_color.currentText())
        settings.setValue(CARLA_KEY_MAIN_REFRESH_INTERVAL, self.ui.sb_main_refresh_interval.value())
        settings.setValue(CARLA_KEY_MAIN_SYSTEM_ICONS,     self.ui.ch_main_system_icons.isChecked())

        # -------------------------------------------------------------------------------------------------------------
        # Canvas

        settings.setValue(CARLA_KEY_CANVAS_THEME,             self.ui.cb_canvas_theme.currentText())
        settings.setValue(CARLA_KEY_CANVAS_SIZE,              self.ui.cb_canvas_size.currentText())
        settings.setValue(CARLA_KEY_CANVAS_USE_BEZIER_LINES,  self.ui.cb_canvas_bezier_lines.isChecked())
        settings.setValue(CARLA_KEY_CANVAS_AUTO_HIDE_GROUPS,  self.ui.cb_canvas_hide_groups.isChecked())
        settings.setValue(CARLA_KEY_CANVAS_AUTO_SELECT_ITEMS, self.ui.cb_canvas_auto_select.isChecked())
        settings.setValue(CARLA_KEY_CANVAS_EYE_CANDY,         self.ui.cb_canvas_eyecandy.isChecked())
        settings.setValue(CARLA_KEY_CANVAS_FANCY_EYE_CANDY,   self.ui.cb_canvas_fancy_eyecandy.isChecked())
        settings.setValue(CARLA_KEY_CANVAS_USE_OPENGL,        self.ui.cb_canvas_use_opengl.isChecked())
        settings.setValue(CARLA_KEY_CANVAS_HQ_ANTIALIASING,   self.ui.cb_canvas_render_hq_aa.isChecked())
        # NOTE: checkState() 0, 1, 2 match their enum variants
        settings.setValue(CARLA_KEY_CANVAS_ANTIALIASING,      self.ui.cb_canvas_render_aa.checkState())
        settings.setValue(CARLA_KEY_CANVAS_FULL_REPAINTS,     self.ui.cb_canvas_full_repaints.isChecked())
        settings.setValue(CARLA_KEY_CANVAS_INLINE_DISPLAYS,   self.ui.cb_canvas_inline_displays.isChecked())

        # -------------------------------------------------------------------------------------------------------------

        settings = QSafeSettings("falkTX", "Carla2")

        # -------------------------------------------------------------------------------------------------------------
        # Main

        settings.setValue(CARLA_KEY_MAIN_EXPERIMENTAL, self.host.experimental)

        # -------------------------------------------------------------------------------------------------------------
        # Engine

        audioDriver = self.ui.cb_engine_audio_driver.currentText()

        if audioDriver and self.host.audioDriverForced is None and not self.host.isPlugin:
            settings.setValue(CARLA_KEY_ENGINE_AUDIO_DRIVER, audioDriver)

        if not self.host.processModeForced:
            # engine sends callback if processMode really changes
            if audioDriver == "JACK":
                self.host.nextProcessMode = self.ui.cb_engine_process_mode_jack.currentIndex()
            else:
                self.host.nextProcessMode = (self.ui.cb_engine_process_mode_other.currentIndex() +
                                             self.PROCESS_MODE_NON_JACK_PADDING)

            settings.setValue(CARLA_KEY_ENGINE_PROCESS_MODE, self.host.nextProcessMode)

        self.host.exportLV2           = self.ui.ch_exp_export_lv2.isChecked()
        self.host.forceStereo         = self.ui.ch_engine_force_stereo.isChecked()
        self.host.resetXruns          = self.ui.cb_engine_reset_xruns.isChecked()
        self.host.maxParameters       = self.ui.sb_engine_max_params.value()
        self.host.manageUIs           = self.ui.ch_engine_manage_uis.isChecked()
        self.host.preferPluginBridges = self.ui.ch_engine_prefer_plugin_bridges.isChecked()
        self.host.preferUIBridges     = self.ui.ch_engine_prefer_ui_bridges.isChecked()
        self.host.showLogs            = self.ui.ch_main_show_logs.isChecked()
        self.host.showPluginBridges   = self.ui.cb_exp_plugin_bridges.isChecked()
        self.host.showWineBridges     = self.ui.ch_exp_wine_bridges.isChecked()
        self.host.uiBridgesTimeout    = self.ui.sb_engine_ui_bridges_timeout.value()
        self.host.uisAlwaysOnTop      = self.ui.ch_engine_uis_always_on_top.isChecked()

        if self.ui.ch_engine_force_stereo.isEnabled():
            settings.setValue(CARLA_KEY_ENGINE_FORCE_STEREO,      self.host.forceStereo)

        settings.setValue(CARLA_KEY_MAIN_SHOW_LOGS,               self.host.showLogs)
        settings.setValue(CARLA_KEY_ENGINE_MAX_PARAMETERS,        self.host.maxParameters)
        settings.setValue(CARLA_KEY_ENGINE_RESET_XRUNS,           self.host.resetXruns)
        settings.setValue(CARLA_KEY_ENGINE_MANAGE_UIS,            self.host.manageUIs)
        settings.setValue(CARLA_KEY_ENGINE_PREFER_PLUGIN_BRIDGES, self.host.preferPluginBridges)
        settings.setValue(CARLA_KEY_ENGINE_PREFER_UI_BRIDGES,     self.host.preferUIBridges)
        settings.setValue(CARLA_KEY_ENGINE_UI_BRIDGES_TIMEOUT,    self.host.uiBridgesTimeout)
        settings.setValue(CARLA_KEY_ENGINE_UIS_ALWAYS_ON_TOP,     self.host.uisAlwaysOnTop)
        settings.setValue(CARLA_KEY_EXPERIMENTAL_EXPORT_LV2,      self.host.exportLV2)
        settings.setValue(CARLA_KEY_EXPERIMENTAL_PLUGIN_BRIDGES,  self.host.showPluginBridges)
        settings.setValue(CARLA_KEY_EXPERIMENTAL_WINE_BRIDGES,    self.host.showWineBridges)

        # -------------------------------------------------------------------------------------------------------------
        # OSC

        settings.setValue(CARLA_KEY_OSC_ENABLED,          self.ui.ch_osc_enable.isChecked())
        settings.setValue(CARLA_KEY_OSC_TCP_PORT_ENABLED, self.ui.group_osc_tcp_port.isChecked())
        settings.setValue(CARLA_KEY_OSC_UDP_PORT_ENABLED, self.ui.group_osc_udp_port.isChecked())
        settings.setValue(CARLA_KEY_OSC_TCP_PORT_RANDOM,  self.ui.rb_osc_tcp_port_random.isChecked())
        settings.setValue(CARLA_KEY_OSC_UDP_PORT_RANDOM,  self.ui.rb_osc_udp_port_random.isChecked())
        settings.setValue(CARLA_KEY_OSC_TCP_PORT_NUMBER,  self.ui.sb_osc_tcp_port_number.value())
        settings.setValue(CARLA_KEY_OSC_UDP_PORT_NUMBER,  self.ui.sb_osc_udp_port_number.value())

        # -------------------------------------------------------------------------------------------------------------
        # File Paths

        audioPaths = []
        midiPaths  = []

        for i in range(self.ui.lw_files_audio.count()):
            audioPaths.append(self.ui.lw_files_audio.item(i).text())

        for i in range(self.ui.lw_files_midi.count()):
            midiPaths.append(self.ui.lw_files_midi.item(i).text())

        self.host.set_engine_option(ENGINE_OPTION_FILE_PATH, FILE_AUDIO, splitter.join(audioPaths))
        self.host.set_engine_option(ENGINE_OPTION_FILE_PATH, FILE_MIDI,  splitter.join(midiPaths))

        settings.setValue(CARLA_KEY_PATHS_AUDIO, audioPaths)
        settings.setValue(CARLA_KEY_PATHS_MIDI,  midiPaths)

        # -------------------------------------------------------------------------------------------------------------
        # Plugin Paths

        ladspas = []
        dssis   = []
        lv2s    = []
        vst2s   = []
        vst3s   = []
        sf2s    = []
        sfzs    = []
        jsfxs   = []

        for i in range(self.ui.lw_ladspa.count()):
            ladspas.append(self.ui.lw_ladspa.item(i).text())

        for i in range(self.ui.lw_dssi.count()):
            dssis.append(self.ui.lw_dssi.item(i).text())

        for i in range(self.ui.lw_lv2.count()):
            lv2s.append(self.ui.lw_lv2.item(i).text())

        for i in range(self.ui.lw_vst.count()):
            vst2s.append(self.ui.lw_vst.item(i).text())

        for i in range(self.ui.lw_vst3.count()):
            vst3s.append(self.ui.lw_vst3.item(i).text())

        for i in range(self.ui.lw_sf2.count()):
            sf2s.append(self.ui.lw_sf2.item(i).text())

        for i in range(self.ui.lw_sfz.count()):
            sfzs.append(self.ui.lw_sfz.item(i).text())

        for i in range(self.ui.lw_jsfx.count()):
            jsfxs.append(self.ui.lw_jsfx.item(i).text())

        self.host.set_engine_option(ENGINE_OPTION_PLUGIN_PATH, PLUGIN_LADSPA, splitter.join(ladspas))
        self.host.set_engine_option(ENGINE_OPTION_PLUGIN_PATH, PLUGIN_DSSI,   splitter.join(dssis))
        self.host.set_engine_option(ENGINE_OPTION_PLUGIN_PATH, PLUGIN_LV2,    splitter.join(lv2s))
        self.host.set_engine_option(ENGINE_OPTION_PLUGIN_PATH, PLUGIN_VST2,   splitter.join(vst2s))
        self.host.set_engine_option(ENGINE_OPTION_PLUGIN_PATH, PLUGIN_VST3,   splitter.join(vst3s))
        self.host.set_engine_option(ENGINE_OPTION_PLUGIN_PATH, PLUGIN_SF2,    splitter.join(sf2s))
        self.host.set_engine_option(ENGINE_OPTION_PLUGIN_PATH, PLUGIN_SFZ,    splitter.join(sfzs))
        self.host.set_engine_option(ENGINE_OPTION_PLUGIN_PATH, PLUGIN_JSFX,   splitter.join(jsfxs))

        settings.setValue(CARLA_KEY_PATHS_LADSPA, ladspas)
        settings.setValue(CARLA_KEY_PATHS_DSSI,   dssis)
        settings.setValue(CARLA_KEY_PATHS_LV2,    lv2s)
        settings.setValue(CARLA_KEY_PATHS_VST2,   vst2s)
        settings.setValue(CARLA_KEY_PATHS_VST3,   vst3s)
        settings.setValue(CARLA_KEY_PATHS_SF2,    sf2s)
        settings.setValue(CARLA_KEY_PATHS_SFZ,    sfzs)
        settings.setValue(CARLA_KEY_PATHS_JSFX,   jsfxs)

        # -------------------------------------------------------------------------------------------------------------
        # Wine

        settings.setValue(CARLA_KEY_WINE_EXECUTABLE, self.ui.le_wine_exec.text())
        settings.setValue(CARLA_KEY_WINE_AUTO_PREFIX, self.ui.cb_wine_prefix_detect.isChecked())
        settings.setValue(CARLA_KEY_WINE_FALLBACK_PREFIX, self.ui.le_wine_prefix_fallback.text())
        settings.setValue(CARLA_KEY_WINE_RT_PRIO_ENABLED, self.ui.group_wine_realtime.isChecked())
        settings.setValue(CARLA_KEY_WINE_BASE_RT_PRIO, self.ui.sb_wine_base_prio.value())
        settings.setValue(CARLA_KEY_WINE_SERVER_RT_PRIO, self.ui.sb_wine_server_prio.value())

        # -------------------------------------------------------------------------------------------------------------
        # Experimental

        settings.setValue(CARLA_KEY_EXPERIMENTAL_JACK_APPS, self.ui.ch_exp_jack_apps.isChecked())
        settings.setValue(CARLA_KEY_EXPERIMENTAL_LOAD_LIB_GLOBAL, self.ui.ch_exp_load_lib_global.isChecked())
        settings.setValue(CARLA_KEY_EXPERIMENTAL_PREVENT_BAD_BEHAVIOUR,
                          self.ui.ch_exp_prevent_bad_behaviour.isChecked())

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot()
    def slot_resetSettings(self):
        currentRow = self.ui.lw_page.currentRow()

        # -------------------------------------------------------------------------------------------------------------
        # Main

        if currentRow == self.TAB_INDEX_MAIN:
            self.ui.le_main_proj_folder.setText(CARLA_DEFAULT_MAIN_PROJECT_FOLDER)
            self.ui.ch_main_theme_pro.setChecked(CARLA_DEFAULT_MAIN_USE_PRO_THEME and
                                                 self.ui.group_main_theme.isEnabled())
            self.ui.cb_main_theme_color.setCurrentIndex(
                self.ui.cb_main_theme_color.findText(CARLA_DEFAULT_MAIN_PRO_THEME_COLOR))
            self.ui.sb_main_refresh_interval.setValue(CARLA_DEFAULT_MAIN_REFRESH_INTERVAL)
            self.ui.ch_main_confirm_exit.setChecked(CARLA_DEFAULT_MAIN_CONFIRM_EXIT)
            self.ui.cb_main_classic_skin_default(CARLA_DEFAULT_MAIN_CLASSIC_SKIN)
            self.ui.ch_main_show_logs.setChecked(CARLA_DEFAULT_MAIN_SHOW_LOGS)

        # -------------------------------------------------------------------------------------------------------------
        # Canvas

        elif currentRow == self.TAB_INDEX_CANVAS:
            self.ui.cb_canvas_theme.setCurrentIndex(self.ui.cb_canvas_theme.findText(CARLA_DEFAULT_CANVAS_THEME))
            self.ui.cb_canvas_size.setCurrentIndex(self.ui.cb_canvas_size.findText(CARLA_DEFAULT_CANVAS_SIZE))
            self.ui.cb_canvas_bezier_lines.setChecked(CARLA_DEFAULT_CANVAS_USE_BEZIER_LINES)
            self.ui.cb_canvas_hide_groups.setChecked(CARLA_DEFAULT_CANVAS_AUTO_HIDE_GROUPS)
            self.ui.cb_canvas_auto_select.setChecked(CARLA_DEFAULT_CANVAS_AUTO_SELECT_ITEMS)
            self.ui.cb_canvas_eyecandy.setChecked(CARLA_DEFAULT_CANVAS_EYE_CANDY and self.ui.cb_canvas_eyecandy.isEnabled())
            self.ui.cb_canvas_render_aa.setCheckState(Qt.PartiallyChecked) # CARLA_DEFAULT_CANVAS_ANTIALIASING
            self.ui.cb_canvas_full_repaints.setChecked(CARLA_DEFAULT_CANVAS_FULL_REPAINTS)

        # -------------------------------------------------------------------------------------------------------------
        # Engine

        elif currentRow == self.TAB_INDEX_ENGINE:
            if not self.host.isPlugin:
                self.ui.cb_engine_audio_driver.setCurrentIndex(0)

            if not self.host.processModeForced:
                if self.ui.cb_engine_audio_driver.currentText() == "JACK":
                    self.ui.cb_engine_process_mode_jack.setCurrentIndex(CARLA_DEFAULT_PROCESS_MODE)
                    self.ui.sw_engine_process_mode.setCurrentIndex(0) # show all modes
                else:
                    self.ui.cb_engine_process_mode_other.setCurrentIndex(CARLA_DEFAULT_PROCESS_MODE -
                                                                         self.PROCESS_MODE_NON_JACK_PADDING)
                    self.ui.sw_engine_process_mode.setCurrentIndex(1) # hide single+multi client modes

            self.ui.sb_engine_max_params.setValue(CARLA_DEFAULT_MAX_PARAMETERS)
            self.ui.cb_engine_reset_xruns.setChecked(CARLA_DEFAULT_RESET_XRUNS)
            self.ui.ch_engine_uis_always_on_top.setChecked(CARLA_DEFAULT_UIS_ALWAYS_ON_TOP)
            self.ui.ch_engine_prefer_ui_bridges.setChecked(CARLA_DEFAULT_PREFER_UI_BRIDGES)
            self.ui.sb_engine_ui_bridges_timeout.setValue(CARLA_DEFAULT_UI_BRIDGES_TIMEOUT)
            self.ui.ch_engine_manage_uis.setChecked(CARLA_DEFAULT_MANAGE_UIS)

        # -------------------------------------------------------------------------------------------------------------
        # OSC

        elif currentRow == self.TAB_INDEX_OSC:
            self.ui.ch_osc_enable.setChecked(CARLA_DEFAULT_OSC_ENABLED)
            self.ui.group_osc_tcp_port.setChecked(CARLA_DEFAULT_OSC_TCP_PORT_ENABLED)
            self.ui.group_osc_udp_port.setChecked(CARLA_DEFAULT_OSC_UDP_PORT_ENABLED)
            self.ui.sb_osc_tcp_port_number.setValue(CARLA_DEFAULT_OSC_TCP_PORT_NUMBER)
            self.ui.sb_osc_udp_port_number.setValue(CARLA_DEFAULT_OSC_UDP_PORT_NUMBER)

            if CARLA_DEFAULT_OSC_TCP_PORT_RANDOM:
                self.ui.rb_osc_tcp_port_specific.setChecked(False)
                self.ui.rb_osc_tcp_port_random.setChecked(True)
            else:
                self.ui.rb_osc_tcp_port_random.setChecked(False)
                self.ui.rb_osc_tcp_port_specific.setChecked(True)

            if CARLA_DEFAULT_OSC_UDP_PORT_RANDOM:
                self.ui.rb_osc_udp_port_specific.setChecked(False)
                self.ui.rb_osc_udp_port_random.setChecked(True)
            else:
                self.ui.rb_osc_udp_port_random.setChecked(False)
                self.ui.rb_osc_udp_port_specific.setChecked(True)

        # -------------------------------------------------------------------------------------------------------------
        # Plugin Paths

        elif currentRow == self.TAB_INDEX_FILEPATHS:
            curIndex = self.ui.tw_filepaths.currentIndex()

            if curIndex == self.FILEPATH_INDEX_AUDIO:
                self.ui.lw_files_audio.clear()

            elif curIndex == self.FILEPATH_INDEX_MIDI:
                self.ui.lw_files_midi.clear()

        # -------------------------------------------------------------------------------------------------------------
        # Plugin Paths

        elif currentRow == self.TAB_INDEX_PLUGINPATHS:
            curIndex = self.ui.tw_paths.currentIndex()

            if curIndex == self.PLUGINPATH_INDEX_LADSPA:
                paths = CARLA_DEFAULT_LADSPA_PATH
                paths.sort()
                self.ui.lw_ladspa.clear()

                for path in paths:
                    if not path:
                        continue
                    self.ui.lw_ladspa.addItem(path)

            elif curIndex == self.PLUGINPATH_INDEX_DSSI:
                paths = CARLA_DEFAULT_DSSI_PATH
                paths.sort()
                self.ui.lw_dssi.clear()

                for path in paths:
                    if not path:
                        continue
                    self.ui.lw_dssi.addItem(path)

            elif curIndex == self.PLUGINPATH_INDEX_LV2:
                paths = CARLA_DEFAULT_LV2_PATH
                paths.sort()
                self.ui.lw_lv2.clear()

                for path in paths:
                    if not path:
                        continue
                    self.ui.lw_lv2.addItem(path)

            elif curIndex == self.PLUGINPATH_INDEX_VST2:
                paths = CARLA_DEFAULT_VST2_PATH
                paths.sort()
                self.ui.lw_vst.clear()

                for path in paths:
                    if not path:
                        continue
                    self.ui.lw_vst.addItem(path)

            elif curIndex == self.PLUGINPATH_INDEX_VST3:
                paths = CARLA_DEFAULT_VST3_PATH
                paths.sort()
                self.ui.lw_vst3.clear()

                for path in paths:
                    if not path:
                        continue
                    self.ui.lw_vst3.addItem(path)

            elif curIndex == self.PLUGINPATH_INDEX_SF2:
                paths = CARLA_DEFAULT_SF2_PATH
                paths.sort()
                self.ui.lw_sf2.clear()

                for path in paths:
                    if not path:
                        continue
                    self.ui.lw_sf2.addItem(path)

            elif curIndex == self.PLUGINPATH_INDEX_SFZ:
                paths = CARLA_DEFAULT_SFZ_PATH
                paths.sort()
                self.ui.lw_sfz.clear()

                for path in paths:
                    if not path:
                        continue
                    self.ui.lw_sfz.addItem(path)

            elif curIndex == self.PLUGINPATH_INDEX_JSFX:
                paths = CARLA_DEFAULT_JSFX_PATH
                paths.sort()
                self.ui.lw_jsfx.clear()

                for path in paths:
                    if not path:
                        continue
                    self.ui.lw_jsfx.addItem(path)

        # -------------------------------------------------------------------------------------------------------------
        # Wine

        elif currentRow == self.TAB_INDEX_WINE:
            self.ui.le_wine_exec.setText(CARLA_DEFAULT_WINE_EXECUTABLE)
            self.ui.cb_wine_prefix_detect.setChecked(CARLA_DEFAULT_WINE_AUTO_PREFIX)
            self.ui.le_wine_prefix_fallback.setText(CARLA_DEFAULT_WINE_FALLBACK_PREFIX)
            self.ui.group_wine_realtime.setChecked(CARLA_DEFAULT_WINE_RT_PRIO_ENABLED)
            self.ui.sb_wine_base_prio.setValue(CARLA_DEFAULT_WINE_BASE_RT_PRIO)
            self.ui.sb_wine_server_prio.setValue(CARLA_DEFAULT_WINE_SERVER_RT_PRIO)

        # -------------------------------------------------------------------------------------------------------------
        # Experimental

        elif currentRow == self.TAB_INDEX_EXPERIMENTAL:
            self.resetExperimentalSettings()

    def resetExperimentalSettings(self):
        # Forever experimental
        self.ui.cb_exp_plugin_bridges.setChecked(CARLA_DEFAULT_EXPERIMENTAL_PLUGIN_BRIDGES)
        self.ui.ch_exp_wine_bridges.setChecked(CARLA_DEFAULT_EXPERIMENTAL_WINE_BRIDGES)
        self.ui.ch_exp_jack_apps.setChecked(CARLA_DEFAULT_EXPERIMENTAL_JACK_APPS)
        self.ui.ch_exp_export_lv2.setChecked(CARLA_DEFAULT_EXPERIMENTAL_LV2_EXPORT)
        self.ui.ch_exp_load_lib_global.setChecked(CARLA_DEFAULT_EXPERIMENTAL_LOAD_LIB_GLOBAL)
        self.ui.ch_exp_prevent_bad_behaviour.setChecked(CARLA_DEFAULT_EXPERIMENTAL_PREVENT_BAD_BEHAVIOUR)

        # Temporary, until stable
        self.ui.ch_main_system_icons.setChecked(CARLA_DEFAULT_MAIN_SYSTEM_ICONS)
        self.ui.cb_canvas_fancy_eyecandy.setChecked(CARLA_DEFAULT_CANVAS_FANCY_EYE_CANDY)
        self.ui.cb_canvas_use_opengl.setChecked(CARLA_DEFAULT_CANVAS_USE_OPENGL and
                                                self.ui.cb_canvas_use_opengl.isEnabled())
        self.ui.cb_canvas_render_hq_aa.setChecked(CARLA_DEFAULT_CANVAS_HQ_ANTIALIASING and
                                                  self.ui.cb_canvas_render_hq_aa.isEnabled())
        self.ui.cb_canvas_inline_displays.setChecked(CARLA_DEFAULT_CANVAS_INLINE_DISPLAYS)
        self.ui.ch_engine_force_stereo.setChecked(CARLA_DEFAULT_FORCE_STEREO)
        self.ui.ch_engine_prefer_plugin_bridges.setChecked(CARLA_DEFAULT_PREFER_PLUGIN_BRIDGES)

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot(bool)
    def slot_enableExperimental(self, toggled):
        if toggled:
            self.ui.lw_page.showRow(self.TAB_INDEX_EXPERIMENTAL)
            if self.ui.ch_exp_wine_bridges.isChecked() and not self.host.isControl:
                self.ui.lw_page.showRow(self.TAB_INDEX_WINE)
        else:
            self.ui.lw_page.hideRow(self.TAB_INDEX_EXPERIMENTAL)
            self.ui.lw_page.hideRow(self.TAB_INDEX_WINE)

    @pyqtSlot(bool)
    def slot_enableWineBridges(self, toggled):
        if toggled and not self.host.isControl:
            self.ui.lw_page.showRow(self.TAB_INDEX_WINE)
        else:
            self.ui.lw_page.hideRow(self.TAB_INDEX_WINE)

    @pyqtSlot(bool)
    def slot_pluginBridgesToggled(self, toggled):
        if not toggled:
            self.ui.ch_exp_wine_bridges.setChecked(False)
            self.ui.ch_engine_prefer_plugin_bridges.setChecked(False)
            self.ui.lw_page.hideRow(self.TAB_INDEX_WINE)

    @pyqtSlot(bool)
    def slot_canvasEyeCandyToggled(self, toggled):
        if not toggled:
            # disable fancy eyecandy too
            self.ui.cb_canvas_fancy_eyecandy.setChecked(False)

    @pyqtSlot(bool)
    def slot_canvasFancyEyeCandyToggled(self, toggled):
        if toggled:
            # make sure normal eyecandy is enabled too
            self.ui.cb_canvas_eyecandy.setChecked(True)

    @pyqtSlot(bool)
    def slot_canvasOpenGLToggled(self, toggled):
        if not toggled:
            # uncheck GL specific option
            self.ui.cb_canvas_render_hq_aa.setChecked(False)

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot()
    def slot_getAndSetProjectPath(self):
        # FIXME
        getAndSetPath(self, self.ui.le_main_proj_folder)

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot()
    def slot_engineAudioDriverChanged(self):
        if self.ui.cb_engine_audio_driver.currentText() == "JACK":
            self.ui.sw_engine_process_mode.setCurrentIndex(0)
        else:
            self.ui.sw_engine_process_mode.setCurrentIndex(1)

    @pyqtSlot()
    def slot_showAudioDriverSettings(self):
        driverIndex = self.ui.cb_engine_audio_driver.currentIndex()
        driverName  = self.ui.cb_engine_audio_driver.currentText()
        DriverSettingsW(self, self.host, driverIndex, driverName).exec_()

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot()
    def slot_addPluginPath(self):
        newPath = QFileDialog.getExistingDirectory(self, self.tr("Add Path"), "", QFileDialog.ShowDirsOnly)

        if not newPath:
            return

        curIndex = self.ui.tw_paths.currentIndex()

        if curIndex == self.PLUGINPATH_INDEX_LADSPA:
            self.ui.lw_ladspa.addItem(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_DSSI:
            self.ui.lw_dssi.addItem(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_LV2:
            self.ui.lw_lv2.addItem(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_VST2:
            self.ui.lw_vst.addItem(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_VST3:
            self.ui.lw_vst3.addItem(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_SF2:
            self.ui.lw_sf2.addItem(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_SFZ:
            self.ui.lw_sfz.addItem(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_JSFX:
            self.ui.lw_jsfx.addItem(newPath)

    @pyqtSlot()
    def slot_removePluginPath(self):
        curIndex = self.ui.tw_paths.currentIndex()

        if curIndex == self.PLUGINPATH_INDEX_LADSPA:
            self.ui.lw_ladspa.takeItem(self.ui.lw_ladspa.currentRow())
        elif curIndex == self.PLUGINPATH_INDEX_DSSI:
            self.ui.lw_dssi.takeItem(self.ui.lw_dssi.currentRow())
        elif curIndex == self.PLUGINPATH_INDEX_LV2:
            self.ui.lw_lv2.takeItem(self.ui.lw_lv2.currentRow())
        elif curIndex == self.PLUGINPATH_INDEX_VST2:
            self.ui.lw_vst.takeItem(self.ui.lw_vst.currentRow())
        elif curIndex == self.PLUGINPATH_INDEX_VST3:
            self.ui.lw_vst3.takeItem(self.ui.lw_vst3.currentRow())
        elif curIndex == self.PLUGINPATH_INDEX_SF2:
            self.ui.lw_sf2.takeItem(self.ui.lw_sf2.currentRow())
        elif curIndex == self.PLUGINPATH_INDEX_SFZ:
            self.ui.lw_sfz.takeItem(self.ui.lw_sfz.currentRow())
        elif curIndex == self.PLUGINPATH_INDEX_JSFX:
            self.ui.lw_jsfx.takeItem(self.ui.lw_jsfx.currentRow())

    @pyqtSlot()
    def slot_changePluginPath(self):
        curIndex = self.ui.tw_paths.currentIndex()

        if curIndex == self.PLUGINPATH_INDEX_LADSPA:
            currentPath = self.ui.lw_ladspa.currentItem().text()
        elif curIndex == self.PLUGINPATH_INDEX_DSSI:
            currentPath = self.ui.lw_dssi.currentItem().text()
        elif curIndex == self.PLUGINPATH_INDEX_LV2:
            currentPath = self.ui.lw_lv2.currentItem().text()
        elif curIndex == self.PLUGINPATH_INDEX_VST2:
            currentPath = self.ui.lw_vst.currentItem().text()
        elif curIndex == self.PLUGINPATH_INDEX_VST3:
            currentPath = self.ui.lw_vst3.currentItem().text()
        elif curIndex == self.PLUGINPATH_INDEX_SF2:
            currentPath = self.ui.lw_sf2.currentItem().text()
        elif curIndex == self.PLUGINPATH_INDEX_SFZ:
            currentPath = self.ui.lw_sfz.currentItem().text()
        elif curIndex == self.PLUGINPATH_INDEX_JSFX:
            currentPath = self.ui.lw_jsfx.currentItem().text()
        else:
            currentPath = ""

        newPath = QFileDialog.getExistingDirectory(self, self.tr("Add Path"), currentPath, QFileDialog.ShowDirsOnly)

        if not newPath:
            return

        if curIndex == self.PLUGINPATH_INDEX_LADSPA:
            self.ui.lw_ladspa.currentItem().setText(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_DSSI:
            self.ui.lw_dssi.currentItem().setText(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_LV2:
            self.ui.lw_lv2.currentItem().setText(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_VST2:
            self.ui.lw_vst.currentItem().setText(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_VST3:
            self.ui.lw_vst3.currentItem().setText(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_SF2:
            self.ui.lw_sf2.currentItem().setText(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_SFZ:
            self.ui.lw_sfz.currentItem().setText(newPath)
        elif curIndex == self.PLUGINPATH_INDEX_JSFX:
            self.ui.lw_jsfx.currentItem().setText(newPath)

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot(int)
    def slot_pluginPathTabChanged(self, index):
        if index == self.PLUGINPATH_INDEX_LADSPA:
            row = self.ui.lw_ladspa.currentRow()
        elif index == self.PLUGINPATH_INDEX_DSSI:
            row = self.ui.lw_dssi.currentRow()
        elif index == self.PLUGINPATH_INDEX_LV2:
            row = self.ui.lw_lv2.currentRow()
        elif index == self.PLUGINPATH_INDEX_VST2:
            row = self.ui.lw_vst.currentRow()
        elif index == self.PLUGINPATH_INDEX_VST3:
            row = self.ui.lw_vst3.currentRow()
        elif index == self.PLUGINPATH_INDEX_SF2:
            row = self.ui.lw_sf2.currentRow()
        elif index == self.PLUGINPATH_INDEX_SFZ:
            row = self.ui.lw_sfz.currentRow()
        elif index == self.PLUGINPATH_INDEX_JSFX:
            row = self.ui.lw_jsfx.currentRow()
        else:
            row = -1

        self.slot_pluginPathRowChanged(row)

    @pyqtSlot(int)
    def slot_pluginPathRowChanged(self, row):
        check = bool(row >= 0)
        self.ui.b_paths_remove.setEnabled(check)
        self.ui.b_paths_change.setEnabled(check)

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot()
    def slot_addFilePath(self):
        newPath = QFileDialog.getExistingDirectory(self, self.tr("Add Path"), "", QFileDialog.ShowDirsOnly)

        if not newPath:
            return

        curIndex = self.ui.tw_filepaths.currentIndex()

        if curIndex == self.FILEPATH_INDEX_AUDIO:
            self.ui.lw_files_audio.addItem(newPath)
        elif curIndex == self.FILEPATH_INDEX_MIDI:
            self.ui.lw_files_midi.addItem(newPath)

    @pyqtSlot()
    def slot_removeFilePath(self):
        curIndex = self.ui.tw_filepaths.currentIndex()

        if curIndex == self.FILEPATH_INDEX_AUDIO:
            self.ui.lw_files_audio.takeItem(self.ui.lw_files_audio.currentRow())
        elif curIndex == self.FILEPATH_INDEX_MIDI:
            self.ui.lw_files_midi.takeItem(self.ui.lw_files_midi.currentRow())

    @pyqtSlot()
    def slot_changeFilePath(self):
        curIndex = self.ui.tw_filepaths.currentIndex()

        if curIndex == self.FILEPATH_INDEX_AUDIO:
            currentPath = self.ui.lw_files_audio.currentItem().text()
        elif curIndex == self.FILEPATH_INDEX_MIDI:
            currentPath = self.ui.lw_files_midi.currentItem().text()
        else:
            currentPath = ""

        newPath = QFileDialog.getExistingDirectory(self, self.tr("Add Path"), currentPath, QFileDialog.ShowDirsOnly)

        if not newPath:
            return

        if curIndex == self.FILEPATH_INDEX_AUDIO:
            self.ui.lw_files_audio.currentItem().setText(newPath)
        elif curIndex == self.FILEPATH_INDEX_MIDI:
            self.ui.lw_files_midi.currentItem().setText(newPath)

    # -----------------------------------------------------------------------------------------------------------------

    @pyqtSlot(int)
    def slot_filePathTabChanged(self, index):
        if index == self.FILEPATH_INDEX_AUDIO:
            row = self.ui.lw_files_audio.currentRow()
        elif index == self.FILEPATH_INDEX_MIDI:
            row = self.ui.lw_files_midi.currentRow()
        else:
            row = -1

        self.slot_filePathRowChanged(row)

    @pyqtSlot(int)
    def slot_filePathRowChanged(self, row):
        check = bool(row >= 0)
        self.ui.b_filepaths_remove.setEnabled(check)
        self.ui.b_filepaths_change.setEnabled(check)

# ---------------------------------------------------------------------------------------------------------------------
# Main

if __name__ == '__main__':
    from carla_app import CarlaApplication
    from carla_host import initHost as _initHost, loadHostSettings as _loadHostSettings
    # pylint: disable=ungrouped-imports
    from carla_shared import handleInitialCommandLineArguments
    # pylint: enable=ungrouped-imports

    initName, libPrefix = handleInitialCommandLineArguments(__file__ if "__file__" in dir() else None)

    _app  = CarlaApplication("Carla2-Settings", libPrefix)
    _host = _initHost("Carla2-Settings", libPrefix, False, False, False)
    _loadHostSettings(_host)

    _gui = CarlaSettingsW(None, _host, True, True)
    _gui.show()

    _app.exit_exec()

# ---------------------------------------------------------------------------------------------------------------------
